"""General views."""

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import cache_page
from esi.decorators import token_required

from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from app_utils.allianceauth import notify_admins
from app_utils.logging import LoggerAddTag

from moonmining import __title__, tasks
from moonmining.app_settings import MOONMINING_ADMIN_NOTIFICATIONS_ENABLED
from moonmining.models import Owner

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("moonmining.basic_access")
def index(request):
    """Render an index view."""
    if request.user.has_perm("moonmining.extractions_access"):
        return redirect("moonmining:extractions")
    return redirect("moonmining:moons")


@permission_required(["moonmining.add_refinery_owner", "moonmining.basic_access"])
@token_required(scopes=Owner.esi_scopes())  # type: ignore
@login_required
def add_owner(request, token):
    """Render view to add an owner."""
    character_ownership = get_object_or_404(
        request.user.character_ownerships.select_related("character"),
        character__character_id=token.character_id,
    )
    try:
        corporation = EveCorporationInfo.objects.get(
            corporation_id=character_ownership.character.corporation_id
        )
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(
            corp_id=character_ownership.character.corporation_id
        )
        corporation.save()

    owner = Owner.objects.update_or_create(
        corporation=corporation,
        defaults={"character_ownership": character_ownership},
    )[0]
    tasks.update_owner.delay(owner.pk)
    messages.success(request, f"Update of refineries started for {owner}.")
    if MOONMINING_ADMIN_NOTIFICATIONS_ENABLED:
        notify_admins(
            message=_(
                "%(corporation)s was added as new owner by %(user)s."
                % {"corporation": owner, "user": request.user}
            ),
            title=f"{__title__}: Owner added: {owner}",
        )
    return redirect("moonmining:index")


@cache_page(3600)
def modal_loader_body(request):
    """Draw the loader body. Useful for showing a spinner while loading a modal."""
    return render(request, "moonmining/modals/loader_body.html")


def tests(request):
    """Render page with JS tests."""
    return render(request, "moonmining/tests.html")



"""Managers."""

# pylint: disable = missing-class-docstring

from collections import namedtuple
from typing import Any, List, Optional, Tuple

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from django.db.models import ExpressionWrapper, F, FloatField, IntegerField, Sum
from django.db.models.functions import Coalesce
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from eveuniverse.managers import EveTypeManager
from eveuniverse.models import EveMoon

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import (
    MOONMINING_REPROCESSING_YIELD,
    MOONMINING_USE_REPROCESS_PRICING,
)
from .constants import EveCategoryId
from .core import CalculatedExtraction
from .helpers import eve_entity_get_or_create_esi_safe

MAX_THREAD_WORKERS = 20
BULK_BATCH_SIZE = 500
logger = LoggerAddTag(get_extension_logger(__name__), __title__)

SurveyProcessResult = namedtuple(
    "SurveyProcessResult", ["moon_name", "success", "error_name"]
)


class EveOreTypeManger(EveTypeManager):
    def get_queryset(self):
        """Return ore types only."""
        return (
            super()
            .get_queryset()
            .select_related("eve_group")
            .filter(published=True)
            .filter(eve_group__eve_category_id=EveCategoryId.ASTEROID)
        )

    def update_current_prices(self, use_process_pricing: Optional[bool] = None):
        """Update current prices for all ores."""
        from .models import EveOreTypeExtras

        if use_process_pricing is None:
            use_process_pricing = MOONMINING_USE_REPROCESS_PRICING

        for obj in self.filter(published=True).select_related("market_price"):
            if use_process_pricing:
                price = obj.calc_refined_value_per_unit(MOONMINING_REPROCESSING_YIELD)
                pricing_method = EveOreTypeExtras.PricingMethod.REPROCESSED_MATERIALS
            else:
                try:
                    price = obj.market_price.average_price
                    pricing_method = EveOreTypeExtras.PricingMethod.EVE_CLIENT
                except ObjectDoesNotExist:
                    price = None
                    pricing_method = EveOreTypeExtras.PricingMethod.UNKNOWN
            EveOreTypeExtras.objects.update_or_create(
                ore_type=obj,
                defaults={"current_price": price, "pricing_method": pricing_method},
            )


class MiningLedgerRecordManager(models.Manager):
    def get_queryset(self):
        """Add calculated values to all object."""
        sum_price = ExpressionWrapper(
            F("quantity") * Coalesce(F("unit_price"), 0.0),
            output_field=FloatField(),
        )
        sum_volume = ExpressionWrapper(
            F("quantity") * F("ore_type__volume"), output_field=IntegerField()
        )
        return (
            super()
            .get_queryset()
            .select_related("ore_type", "ore_type__extras")
            .annotate(unit_price=F("ore_type__extras__current_price"))
            .annotate(total_price=Sum(sum_price, distinct=True))
            .annotate(total_volume=Sum(sum_volume, distinct=True))
        )


class MoonQuerySet(models.QuerySet):
    def selected_related_defaults(self) -> models.QuerySet:
        """Apply default select_related."""
        return self.select_related(
            "eve_moon",
            "eve_moon__eve_planet__eve_solar_system",
            "eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
            "refinery",
            "refinery__eve_type",
            "refinery__owner",
            "refinery__owner__corporation",
            "refinery__owner__corporation__alliance",
            "label",
        )


class MoonManagerBase(models.Manager):
    def update_moons_from_survey(self, scans: str, user: Optional[User] = None) -> bool:
        """Update moons from survey input.

        Args:
            scans: raw text input from user containing moon survey data
            user: (optional) user who submitted the data
        """
        surveys, error_name = self._parse_scans(scans)
        if surveys:
            process_results, success = self._process_surveys(surveys, user)
        else:
            process_results = None
            success = False

        if user:
            success = self._send_survey_process_report_to_user(
                process_results, error_name, user, success
            )
        return success

    @staticmethod
    def _parse_scans(scans: str) -> tuple:
        surveys = []
        try:
            lines = scans.split("\n")
            lines_ = []
            for line in lines:
                line = line.strip("\r").split("\t")
                lines_.append(line)
            lines = lines_

            # Find all groups of scans.
            if len(lines[0]) == 0 or lines[0][0] == "Moon":
                lines = lines[1:]

            sub_lists = MoonManagerBase._find_lines_that_start_a_scan(lines)

            # Separate out individual surveys
            for i, _obj in enumerate(sub_lists):
                # The First List
                if i == 0:
                    if i + 2 > len(sub_lists):
                        surveys.append(lines[sub_lists[i] :])
                    else:
                        surveys.append(lines[sub_lists[i] : sub_lists[i + 1]])
                else:
                    if i + 2 > len(sub_lists):
                        surveys.append(lines[sub_lists[i] :])
                    else:
                        surveys.append(lines[sub_lists[i] : sub_lists[i + 1]])

        except (TypeError, ValueError, KeyError, AttributeError, IndexError) as ex:
            logger.warning(
                "An issue occurred while trying to parse the surveys", exc_info=True
            )
            error_name = type(ex).__name__

        else:
            error_name = ""
        return surveys, error_name

    @staticmethod
    def _find_lines_that_start_a_scan(lines):
        sub_lists = []
        for line in lines:
            if line[0] == "":
                pass
            else:
                sub_lists.append(lines.index(line))
        return sub_lists

    def _process_surveys(
        self, surveys: list, user: Optional[User]
    ) -> Tuple[List[SurveyProcessResult], bool]:
        from .models import Moon

        overall_success = True
        process_results = []
        for survey in surveys:
            try:
                moon: Moon = self._get_or_create_from_survey(survey)
                moon_products = self._extract_moon_products(survey, moon)
                moon.update_products(moon_products, updated_by=user)
                logger.info("Added moon survey for %s", moon.name)

            except Exception as ex:  # pylint: disable = broad-exception-caught
                # FIXME: Reduce broad exception
                logger.warning(
                    "An issue occurred while processing the following moon survey: %s",
                    survey,
                    exc_info=True,
                )
                error_name = type(ex).__name__
                overall_success = success = False
                moon = None
            else:
                success = True
                error_name = None

            process_results.append(
                SurveyProcessResult(
                    moon_name=moon.name if moon else "",
                    success=success,
                    error_name=error_name,
                )
            )
        return process_results, overall_success

    def _extract_moon_products(self, survey, moon):
        from .models import EveOreType, MoonProduct

        moon_products = []
        survey = survey[1:]
        for product_data in survey:
            # Trim off the empty index at the front
            product_data = product_data[1:]
            ore_type = EveOreType.objects.get_or_create_esi(id=product_data[2])[0]
            moon_products.append(
                MoonProduct(moon=moon, amount=product_data[1], ore_type=ore_type)
            )

        return moon_products

    def _get_or_create_from_survey(self, survey):
        moon_id = survey[1][6]
        eve_moon = EveMoon.objects.get_or_create_esi(id=moon_id)[0]
        moon = self.get_or_create(eve_moon=eve_moon)[0]
        return moon

    @staticmethod
    def _send_survey_process_report_to_user(
        process_results: Optional[List[SurveyProcessResult]],
        error_name: str,
        user: User,
        success: bool,
    ) -> bool:
        message = "We have completed processing your moon survey input:\n\n"
        if process_results:
            for num, process_result in enumerate(process_results):
                moon_name = process_result.moon_name
                if process_result.success:
                    status = "OK"
                    error_name = ""
                else:
                    status = "FAILED"
                    success = False
                    error_name = f"- {process_result.error_name}"
                message += f"#{num + 1}: {moon_name}: {status} {error_name}\n"
        else:
            message += "\nProcessing failed"

        title_detail = "OK" if success else "FAILED"
        notify(
            user=user,
            title=_(f"Moon survey input processing results: {title_detail}"),
            message=message,
            level="success" if success else "danger",
        )
        return success


MoonManager = MoonManagerBase.from_queryset(MoonQuerySet)


class ExtractionQuerySet(models.QuerySet):
    def selected_related_defaults(self) -> models.QuerySet:
        """Apply default select related."""
        return self.select_related(
            "refinery",
            "refinery__moon",
            "refinery__moon__eve_moon",
            "refinery__owner",
            "refinery__owner__corporation",
            "refinery__owner__corporation__alliance",
            "refinery__moon__label",
        )

    def update_status(self):
        """Update status of given extractions according to current time."""
        self.exclude(
            status__in=[self.model.Status.READY, self.model.Status.CANCELED]
        ).filter(
            chunk_arrival_at__lte=now(),
            auto_fracture_at__gt=now(),
        ).update(
            status=self.model.Status.READY
        )
        self.exclude(
            status__in=[self.model.Status.COMPLETED, self.model.Status.CANCELED]
        ).filter(auto_fracture_at__lte=now()).update(status=self.model.Status.COMPLETED)

    def annotate_volume(self) -> models.QuerySet:
        """Add volume of all products"""
        return self.annotate(volume=Sum("products__volume"))


class ExtractionManagerBase(models.Manager):
    def update_from_calculated(self, calculated: CalculatedExtraction) -> bool:
        """Update an extraction object from related calculated extraction
        when there is new information.

        Return True when updated, else False.
        """
        from .models import EveOreType, ExtractionProduct

        try:
            extraction = self._find_matching_extraction(calculated)
        except self.model.DoesNotExist:
            logger.debug("%s: Could not find matching extraction", calculated)
            return False

        needs_update, status_changed = self._calc_update_need_and_status_change(
            calculated, extraction
        )

        updated = False
        if needs_update:
            extraction.save()
            updated = True

        if calculated.products and (status_changed or not extraction.products.exists()):
            # preload eve ore types before transaction starts
            EveOreType.objects.bulk_get_or_create_esi(
                ids=[product.ore_type_id for product in calculated.products]
            )
            products = [
                ExtractionProduct(
                    extraction=extraction,
                    ore_type_id=product.ore_type_id,
                    volume=product.volume,
                )
                for product in calculated.products
            ]
            with transaction.atomic():
                ExtractionProduct.objects.filter(extraction=extraction).delete()
                ExtractionProduct.objects.bulk_create(
                    products, batch_size=BULK_BATCH_SIZE
                )
            extraction.update_calculated_properties()
            updated = True
        return updated

    def _find_matching_extraction(self, calculated: CalculatedExtraction) -> Any:
        if calculated.chunk_arrival_at:
            return self.get(
                refinery_id=calculated.refinery_id,
                chunk_arrival_at=calculated.chunk_arrival_at,
            )

        if calculated.auto_fracture_at:
            return self.get(
                refinery_id=calculated.refinery_id,
                auto_fracture_at=calculated.auto_fracture_at,
            )

        logger.debug(
            "%s: Not enough data to search for matching extraction", calculated
        )
        raise self.model.DoesNotExist()

    def _calc_update_need_and_status_change(self, calculated, extraction):
        needs_update = False
        if calculated.canceled_at and not extraction.canceled_at:
            extraction.canceled_at = calculated.canceled_at
            needs_update = True

        if calculated.canceled_by and not extraction.canceled_by:
            extraction.canceled_by = eve_entity_get_or_create_esi_safe(
                calculated.canceled_by
            )
            needs_update = True

        if calculated.canceled_by and not extraction.canceled_by:
            extraction.canceled_by = eve_entity_get_or_create_esi_safe(
                calculated.canceled_by
            )
            needs_update = True

        if calculated.fractured_by and not extraction.fractured_by:
            extraction.fractured_by = eve_entity_get_or_create_esi_safe(
                calculated.fractured_by
            )
            needs_update = True

        if calculated.fractured_at and not extraction.fractured_at:
            extraction.fractured_at = calculated.fractured_at
            needs_update = True

        if self.model.Status.from_calculated(calculated) != extraction.status:
            extraction.status = self.model.Status.from_calculated(calculated)
            needs_update = True
            status_changed = True
        else:
            status_changed = False

        if calculated.started_by and not extraction.started_by:
            extraction.started_by = eve_entity_get_or_create_esi_safe(
                calculated.started_by
            )
            needs_update = True
        return needs_update, status_changed


ExtractionManager = ExtractionManagerBase.from_queryset(ExtractionQuerySet)


class RefineryManager(models.Manager):
    def ids(self) -> set:
        """Return IDs of this queryset."""
        return set(self.values_list("id", flat=True))



"""Moon views."""

from enum import Enum
from typing import Union

from django_datatables_view.base_datatable_view import BaseDatatableView

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.mixins import LoginRequiredMixin, PermissionRequiredMixin
from django.contrib.auth.models import User
from django.db import models
from django.db.models import Case, F, OuterRef, Q, QuerySet, Subquery, Value, When
from django.db.models.functions import Concat
from django.http import HttpRequest, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.evelinks import dotlan
from app_utils.views import link_html

from moonmining import helpers, tasks
from moonmining.app_settings import (
    MOONMINING_REPROCESSING_YIELD,
    MOONMINING_USE_REPROCESS_PRICING,
    MOONMINING_VOLUME_PER_MONTH,
)
from moonmining.forms import MoonScanForm
from moonmining.helpers import user_perms_lookup
from moonmining.models import Extraction, Moon, MoonProduct
from moonmining.views._helpers import moon_details_button_html
from moonmining.views.extractions import extraction_details_button_html


class MoonsCategory(str, helpers.EnumToDict, Enum):
    """A category defining which moons to show in data view."""

    ALL = "all_moons"
    UPLOADS = "uploads"
    OURS = "our_moons"


# pylint: disable = too-many-ancestors
class MoonListJson(PermissionRequiredMixin, LoginRequiredMixin, BaseDatatableView):
    """A datatable view for rendering a moons list."""

    model = Moon
    permission_required = "moonmining.basic_access"
    columns = [
        "id",
        "moon_name",
        "rarity_class_str",
        "refinery",
        "labels",
        "solar_system_link",
        "location_html",
        "region_name",
        "constellation_name",
        "value",
        "details",
        "has_refinery_str",
        "has_extraction_str",
        "solar_system_name",
        "corporation_name",
        "alliance_name",
        "has_refinery",
        "label_name",
        "ore_type_dummy",
    ]

    # define column names that will be used in sorting
    # order is important and should be same as order of columns
    # displayed by datatables. For non sortable columns use empty
    # value like ''
    order_columns = [
        "eve_moon__name",
        "eve_moon__eve_planet__eve_solar_system__name",
        "eve_moon__eve_planet__eve_solar_system__eve_constellation__name",
        "refinery__name",
        "",
        "value",
        "",
    ]

    def get_initial_queryset(self) -> QuerySet:
        return self.initial_queryset(
            category=self.kwargs["category"], user=self.request.user
        )

    @classmethod
    def initial_queryset(cls, category: str, user: User) -> QuerySet:
        """Return initial queryset."""
        current_extraction_qs = Extraction.objects.filter(
            refinery__moon=OuterRef("pk"),
            status__in=[Extraction.Status.STARTED, Extraction.Status.READY],
        )
        moon_query = (
            Moon.objects.selected_related_defaults()
            .annotate(extraction_pk=Subquery(current_extraction_qs.values("pk")[:1]))
            .annotate(
                has_refinery=Case(
                    When(refinery__isnull=True, then=Value(False)), default=Value(True)
                )
            )
            .annotate(
                has_refinery_str=Case(
                    When(has_refinery=False, then=Value("no")), default=Value("yes")
                )
            )
            .annotate(
                has_extraction=Case(
                    When(extraction_pk__isnull=True, then=Value(False)),
                    default=Value(True),
                )
            )
            .annotate(
                has_extraction_str=Case(
                    When(has_extraction=False, then=Value("no")), default=Value("yes")
                )
            )
            .annotate(
                rarity_class_str=Concat(
                    Value("R"), F("rarity_class"), output_field=models.CharField()
                )
            )
        )

        moons_category = MoonsCategory(category)
        if moons_category is MoonsCategory.ALL and user.has_perm(
            "moonmining.view_all_moons"
        ):
            return moon_query

        if (
            moons_category is MoonsCategory.OURS
            and user.has_perm("moonmining.extractions_access")
            or user.has_perm("moonmining.view_all_moons")
        ):
            return moon_query.filter(refinery__isnull=False)

        if moons_category is MoonsCategory.UPLOADS and user.has_perm(
            "moonmining.upload_moon_scan"
        ):
            return moon_query.filter(products_updated_by=user)

        return Moon.objects.none()

    def filter_queryset(self, qs: QuerySet) -> QuerySet:
        """use parameters passed in GET request to filter queryset"""

        qs = self._apply_search_filter(
            qs, 7, "eve_moon__eve_planet__eve_solar_system__name"
        )
        qs = self._apply_search_filter(qs, 8, "has_refinery_str")
        qs = self._apply_search_filter(
            qs, 9, "refinery__owner__corporation__corporation_name"
        )
        qs = self._apply_search_filter(
            qs, 10, "refinery__owner__corporation__alliance__alliance_name"
        )
        qs = self._apply_search_filter(qs, 11, "rarity_class_str")
        qs = self._apply_search_filter(qs, 12, "has_extraction_str")
        qs = self._apply_search_filter(
            qs, 13, "eve_moon__eve_planet__eve_solar_system__eve_constellation__name"
        )
        qs = self._apply_search_filter(qs, 14, "label__name")
        qs = self._apply_search_filter(
            qs,
            15,
            "eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region__name",
        )

        if ore_type_name := self.request.GET.get("columns[16][search][value]"):
            if ore_type_name := ore_type_name.removeprefix("^").removesuffix("$"):
                qs = qs.filter(products__ore_type__name__in=[ore_type_name])

        if search := self.request.GET.get("search[value]", None):
            qs = qs.filter(
                Q(eve_moon__name__istartswith=search)
                | Q(refinery__name__istartswith=search)
            )
        return qs

        # qs = self._apply_search_filter(qs, 4, "user__profile__state__name")
        # qs = self._apply_search_filter(qs, 6, "character__alliance_name")
        # qs = self._apply_search_filter(qs, 7, "character__corporation_name")
        # qs = self._apply_search_filter(
        #     qs, 8, "user__profile__main_character__alliance_name"
        # )
        # qs = self._apply_search_filter(
        #     qs, 9, "user__profile__main_character__corporation_name"
        # )
        # qs = self._apply_search_filter(
        #     qs, 10, "user__profile__main_character__character_name"
        # )
        # qs = self._apply_search_filter(qs, 11, "unregistered")

        # return qs

    def _apply_search_filter(
        self, qs: QuerySet, column_num: int, field: str
    ) -> QuerySet:
        my_filter = self.request.GET.get(f"columns[{column_num}][search][value]", None)
        if not my_filter:
            return qs

        if self.request.GET.get(f"columns[{column_num}][search][regex]", False):
            kwargs = {f"{field}__iregex": my_filter}
        else:
            kwargs = {f"{field}__istartswith": my_filter}
        return qs.filter(**kwargs)

    # pylint: disable = too-many-return-statements
    def render_column(self, row, column) -> Union[str, dict]:
        if column == "id":
            return row.pk

        if column == "moon_name":
            return row.name

        if result := self._render_location(row, column):
            return result

        if column == "labels":
            return row.labels_html()

        if column == "label_name":
            return row.label.name if row.label else ""

        if column == "details":
            return self._render_details(row)

        if result := self._render_refinery(row, column):
            return result

        if column == "ore_type_dummy":
            return ""

        return super().render_column(row, column)

    def _render_location(self, row, column):
        solar_system = row.eve_moon.eve_planet.eve_solar_system
        if solar_system.is_high_sec:
            sec_class = "text-high-sec"
        elif solar_system.is_low_sec:
            sec_class = "text-low-sec"
        else:
            sec_class = "text-null-sec"

        solar_system_link = format_html(
            '{}&nbsp;<span class="{}">{}</span>',
            link_html(dotlan.solar_system_url(solar_system.name), solar_system.name),
            sec_class,
            round(solar_system.security_status, 1),
        )
        constellation = row.eve_moon.eve_planet.eve_solar_system.eve_constellation
        region = constellation.eve_region
        location_html = format_html(
            "{}<br><em>{}</em>", constellation.name, region.name
        )
        if column == "solar_system_name":
            return solar_system.name

        if column == "solar_system_link":
            return solar_system_link

        if column == "location_html":
            return location_html

        if column == "region_name":
            return region.name

        if column == "constellation_name":
            return constellation.name

        return None

    def _render_details(self, row):
        details_html = ""
        if self.request.user.has_perm("moonmining.extractions_access"):
            details_html = (
                extraction_details_button_html(row.extraction_pk) + " "
                if row.extraction_pk
                else ""
            )
        details_html += moon_details_button_html(row)
        return details_html

    def _render_refinery(self, row, column) -> Union[str, dict]:
        if row.has_refinery:
            refinery = row.refinery
            refinery_html = refinery.name_html()
            refinery_name = refinery.name
            corporation_name = refinery.owner.name
            alliance_name = refinery.owner.alliance_name
        else:
            refinery_html = "?"
            refinery_name = ""
            corporation_name = alliance_name = ""

        if column == "corporation_name":
            return corporation_name

        if column == "alliance_name":
            return alliance_name

        if column == "refinery":
            return {"display": refinery_html, "sort": refinery_name}

        return ""


@login_required
@permission_required("moonmining.basic_access")
def moons_fdd_data(request: HttpRequest, category: str) -> JsonResponse:
    """Provide lists for drop down fields."""
    qs = MoonListJson.initial_queryset(category=category, user=request.user)
    result = {}
    if columns := request.GET.get("columns"):
        for column in columns.split(","):
            options = _calc_options(request, qs, column)
            result[column] = sorted(list(set(options)), key=str.casefold)
    return JsonResponse(result, safe=False)


# pylint: disable = too-many-return-statements
def _calc_options(request: HttpRequest, qs: QuerySet, column: str) -> QuerySet:
    if column == "alliance_name":
        return qs.exclude(
            refinery__owner__corporation__alliance__isnull=True,
        ).values_list(
            "refinery__owner__corporation__alliance__alliance_name", flat=True
        )

    if column == "corporation_name":
        return qs.exclude(refinery__isnull=True).values_list(
            "refinery__owner__corporation__corporation_name", flat=True
        )

    if column == "region_name":
        return qs.values_list(
            "eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region__name",
            flat=True,
        )

    if column == "constellation_name":
        return qs.values_list(
            "eve_moon__eve_planet__eve_solar_system__eve_constellation__name",
            flat=True,
        )

    if column == "solar_system_name":
        return qs.values_list(
            "eve_moon__eve_planet__eve_solar_system__name",
            flat=True,
        )

    if column == "rarity_class_str":
        return qs.values_list("rarity_class_str", flat=True)

    if column == "label_name":
        return qs.exclude(label__isnull=True).values_list("label__name", flat=True)

    if column == "has_refinery_str":
        return qs.values_list("has_refinery_str", flat=True)

    if column == "has_extraction_str":
        if request.user.has_perm("moonmining.extractions_access"):
            return qs.values_list("has_extraction_str", flat=True)
        return []

    if column == "ore_type_dummy":
        moon_ids = set(qs.values_list("eve_moon_id", flat=True))
        return (
            MoonProduct.objects.filter(moon_id__in=moon_ids)
            .select_related("ore_type")
            .values_list("ore_type__name", flat=True)
        )

    return [f"** ERROR: No options defined for column name '{column}' **"]


@login_required()
@permission_required("moonmining.basic_access")
def moons(request):
    """Render moons page."""
    user_perms = user_perms_lookup(
        request.user, ["moonmining.extractions_access", "moonmining.view_all_moons"]
    )
    context = {
        "page_title": _("Moons"),
        "MoonsCategory": MoonsCategory.to_dict(),
        "use_reprocess_pricing": MOONMINING_USE_REPROCESS_PRICING,
        "reprocessing_yield": MOONMINING_REPROCESSING_YIELD * 100,
        "total_volume_per_month": MOONMINING_VOLUME_PER_MONTH / 1000000,
        "user_perms": user_perms,
    }
    return render(request, "moonmining/moons.html", context)


@login_required
@permission_required("moonmining.basic_access")
def moon_details(request, moon_pk: int):
    """Render moon details page."""
    moon = get_object_or_404(Moon.objects.selected_related_defaults(), pk=moon_pk)
    context = {
        "page_title": moon.name,
        "moon": moon,
        "use_reprocess_pricing": MOONMINING_USE_REPROCESS_PRICING,
        "reprocessing_yield": MOONMINING_REPROCESSING_YIELD * 100,
        "total_volume_per_month": MOONMINING_VOLUME_PER_MONTH / 1000000,
    }
    if request.GET.get("new_page"):
        context["title"] = _("Moon")
        context["content_file"] = "moonmining/partials/moon_details.html"
        return render(request, "moonmining/_generic_modal_page.html", context)
    return render(request, "moonmining/modals/moon_details.html", context)


@permission_required(["moonmining.basic_access", "moonmining.upload_moon_scan"])
@login_required()
def upload_survey(request):
    """Render upload survey page."""
    context = {"page_title": _("Upload Moon Surveys")}
    if request.method == "POST":
        form = MoonScanForm(request.POST)
        if form.is_valid():
            scans = request.POST["scan"]
            tasks.process_survey_input.delay(scans, request.user.pk)
            messages.success(
                request,
                _(
                    "Your scan has been submitted for processing. "
                    "You will receive a notification once processing is complete."
                ),
            )
        else:
            messages.error(
                request,
                _(
                    "Oh No! Something went wrong with your moon scan submission. "
                    "Please try again."
                ),
            )
        return redirect("moonmining:moons")
    return render(request, "moonmining/modals/upload_survey.html", context=context)



"""Extraction views."""

import datetime as dt
from enum import Enum

from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import (
    ExpressionWrapper,
    F,
    FloatField,
    IntegerField,
    QuerySet,
    Sum,
    Value,
)
from django.db.models.functions import Coalesce
from django.http import HttpRequest, JsonResponse
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.html import format_html
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.evelinks import dotlan
from app_utils.views import link_html, yesno_str

from moonmining import helpers
from moonmining.app_settings import (
    MOONMINING_COMPLETED_EXTRACTIONS_HOURS_UNTIL_STALE,
    MOONMINING_REPROCESSING_YIELD,
    MOONMINING_USE_REPROCESS_PRICING,
    MOONMINING_VOLUME_PER_MONTH,
)
from moonmining.constants import DATE_FORMAT, DATETIME_FORMAT
from moonmining.models import Extraction
from moonmining.views._helpers import (
    fontawesome_modal_button_html,
    moon_details_button_html,
)


class ExtractionsCategory(str, helpers.EnumToDict, Enum):
    """A category defining which extractions to show in data view."""

    UPCOMING = "upcoming"
    PAST = "past"


@login_required
@permission_required(["moonmining.extractions_access", "moonmining.basic_access"])
def extractions(request):
    """Render extractions page."""
    context = {
        "page_title": _("Extractions"),
        "ExtractionsCategory": ExtractionsCategory.to_dict(),
        "ExtractionsStatus": Extraction.Status,
        "use_reprocess_pricing": MOONMINING_USE_REPROCESS_PRICING,
        "reprocessing_yield": MOONMINING_REPROCESSING_YIELD * 100,
        "total_volume_per_month": MOONMINING_VOLUME_PER_MONTH / 1000000,
        "stale_hours": MOONMINING_COMPLETED_EXTRACTIONS_HOURS_UNTIL_STALE,
    }
    return render(request, "moonmining/extractions.html", context)


def extraction_ledger_button_html(extraction: Extraction) -> str:
    """Return HTML to render extraction ledger button."""
    new_var = fontawesome_modal_button_html(
        modal_id="modalExtractionLedger",
        fa_code="fas fa-table",
        ajax_url=reverse("moonmining:extraction_ledger", args=[extraction.pk]),
        tooltip="Extraction ledger",
    )
    return new_var


def extraction_details_button_html(extraction_pk: int) -> str:
    """Return HTML to render extraction details button."""
    html = fontawesome_modal_button_html(
        modal_id="modalExtractionDetails",
        fa_code="fas fa-hammer",
        ajax_url=reverse("moonmining:extraction_details", args=[extraction_pk]),
        tooltip=_("Extraction details"),
    )
    return html


@login_required
@permission_required(["moonmining.extractions_access", "moonmining.basic_access"])
def extractions_data(request: HttpRequest, category: str):
    """Render extraction data."""
    data = []
    can_see_ledger = request.user.has_perm("moonmining.view_moon_ledgers")
    extractions_qs = _calc_extractions_qs(ExtractionsCategory(category))
    for extraction in extractions_qs:
        moon = extraction.refinery.moon
        moon_name = str(moon)
        refinery_name = str(extraction.refinery.name)
        solar_system = moon.eve_moon.eve_planet.eve_solar_system
        location = format_html(
            "{}<br><i>{}</i>",
            link_html(dotlan.solar_system_url(solar_system.name), moon_name),
            solar_system.eve_constellation.eve_region.name,
        )

        if (
            extraction.status == Extraction.Status.COMPLETED
            and extraction.ledger.exists()
        ):
            mined_value = extraction.ledger.aggregate(Sum(F("total_price")))[
                "total_price__sum"
            ]
            actions_html = (
                extraction_ledger_button_html(extraction) + "&nbsp;"
                if can_see_ledger
                else ""
            )
        else:
            actions_html = ""
            mined_value = None

        actions_html += extraction_details_button_html(extraction.pk)
        actions_html += "&nbsp;" + moon_details_button_html(extraction.refinery.moon)
        status_html = format_html(
            "{}<br>{}",
            extraction.chunk_arrival_at.strftime(DATETIME_FORMAT),
            extraction.status_enum.bootstrap_tag_html,
        )
        data.append(
            {
                "id": extraction.pk,
                "chunk_arrival_at": {
                    "display": status_html,
                    "sort": extraction.chunk_arrival_at,
                },
                "refinery": {
                    "display": extraction.refinery.name_html(),
                    "sort": refinery_name,
                },
                "location": {
                    "display": location,
                    "sort": moon_name,
                },
                "labels": moon.labels_html(),
                "volume": extraction.volume,
                "value": extraction.value if extraction.value else None,
                "mined_value": mined_value,
                "details": actions_html,
                "corporation_name": extraction.refinery.owner.name,
                "alliance_name": extraction.refinery.owner.alliance_name,
                "moon_name": moon_name,
                "region_name": solar_system.eve_constellation.eve_region.name,
                "constellation_name": solar_system.eve_constellation.name,
                "rarity_class": moon.get_rarity_class_display(),
                "is_jackpot_str": yesno_str(extraction.is_jackpot),
                "is_ready": extraction.chunk_arrival_at <= now(),
                "status": extraction.status,
                "status_str": Extraction.Status(extraction.status).label,
            }
        )
    return JsonResponse(data, safe=False)


def _calc_extractions_qs(category: ExtractionsCategory) -> QuerySet[Extraction]:
    stale_cutoff = now() - dt.timedelta(
        hours=MOONMINING_COMPLETED_EXTRACTIONS_HOURS_UNTIL_STALE
    )
    extractions_qs = (
        Extraction.objects.exclude(refinery__moon__isnull=True)
        .annotate_volume()
        .selected_related_defaults()
        .select_related(
            "refinery__moon__eve_moon__eve_planet__eve_solar_system",
            "refinery__moon__eve_moon__eve_planet__eve_solar_system__eve_constellation",
            "refinery__moon__eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
        )
    )
    if category is ExtractionsCategory.UPCOMING:
        extractions_qs = extractions_qs.filter(
            auto_fracture_at__gte=stale_cutoff
        ).exclude(status=Extraction.Status.CANCELED)

    elif category is ExtractionsCategory.PAST:
        extractions_qs = extractions_qs.filter(
            auto_fracture_at__lt=stale_cutoff
        ) | extractions_qs.filter(status=Extraction.Status.CANCELED)

    else:
        extractions_qs = Extraction.objects.none()
    return extractions_qs


@login_required
@permission_required(["moonmining.extractions_access", "moonmining.basic_access"])
def extraction_details(request, extraction_pk: int):
    """Render a details view for an extraction."""
    extraction = get_object_or_404(
        Extraction.objects.annotate_volume().select_related(
            "refinery",
            "refinery__moon",
            "refinery__moon__eve_moon",
            "refinery__moon__eve_moon__eve_planet__eve_solar_system",
            "refinery__moon__eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
            "canceled_by",
            "fractured_by",
            "started_by",
        ),
        pk=extraction_pk,
    )
    context = {
        "page_title": (
            f"{extraction.refinery.moon} "
            f"| {extraction.chunk_arrival_at.strftime(DATE_FORMAT)}"
        ),
        "extraction": extraction,
    }
    if request.GET.get("new_page"):
        context["title"] = _("Extraction")
        context["content_file"] = "moonmining/partials/extraction_details.html"
        return render(request, "moonmining/_generic_modal_page.html", context)

    return render(request, "moonmining/modals/extraction_details.html", context)


@login_required
@permission_required(
    [
        "moonmining.extractions_access",
        "moonmining.basic_access",
        "moonmining.view_moon_ledgers",
    ]
)
def extraction_ledger(request, extraction_pk: int):
    """Render extraction ledger page."""
    extraction = get_object_or_404(
        Extraction.objects.all().select_related(
            "refinery",
            "refinery__moon",
            "refinery__moon__eve_moon__eve_planet__eve_solar_system",
            "refinery__moon__eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
        ),
        pk=extraction_pk,
    )
    ledger = extraction.ledger.select_related(
        "character", "corporation", "user__profile__main_character", "ore_type"
    )
    total_value = ledger.aggregate(Sum(F("total_price")))["total_price__sum"]
    total_volume = ledger.aggregate(Sum(F("total_volume")))["total_volume__sum"]
    sum_price = ExpressionWrapper(
        F("quantity") * Coalesce(F("unit_price"), 0), output_field=FloatField()
    )
    sum_volume = ExpressionWrapper(
        F("quantity") * F("ore_type__volume"), output_field=IntegerField()
    )
    character_totals = (
        ledger.values(
            character_name=F("character__name"),
            main_name=F("user__profile__main_character__character_name"),
            corporation_name=F("user__profile__main_character__corporation_name"),
        )
        .annotate(character_total_price=Sum(sum_price, distinct=True))
        .annotate(character_total_volume=Sum(sum_volume, distinct=True))
        .annotate(
            character_percent_value=ExpressionWrapper(
                F("character_total_price") / Value(total_value) * Value(100),
                output_field=IntegerField(),
            )
        )
        .annotate(
            character_percent_volume=F("character_total_volume")
            / Value(total_volume)
            * Value(100)
        )
    )
    context = {
        "page_title": (
            f"{extraction.refinery.moon} "
            f"| {extraction.chunk_arrival_at.strftime(DATE_FORMAT)}"
        ),
        "extraction": extraction,
        "total_value": total_value,
        "total_volume": total_volume,
        "ledger": ledger,
        "character_totals": character_totals,
    }
    if request.GET.get("new_page"):
        context["title"] = _("Extraction Ledger")
        context["content_file"] = "moonmining/partials/extraction_ledger.html"
        return render(request, "moonmining/_generic_modal_page.html", context)
    return render(request, "moonmining/modals/extraction_ledger.html", context)



from eveuniverse.models import EveMarketPrice, EveType

from app_utils.testing import NoSocketsTestCase

from moonmining.constants import EveTypeId
from moonmining.core import CalculatedExtraction
from moonmining.models import EveOreType, OreQualityClass
from moonmining.tests.testdata.factories import (
    Extraction,
    ExtractionFactory,
    RefineryFactory,
)
from moonmining.tests.testdata.load_allianceauth import load_allianceauth
from moonmining.tests.testdata.load_eveuniverse import load_eveuniverse


class TestEveOreTypeCalcRefinedValues(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def setUp(self) -> None:
        self.cinnebar = EveOreType.objects.get(id=45506)
        tungsten = EveType.objects.get(id=16637)
        mercury = EveType.objects.get(id=16646)
        evaporite_deposits = EveType.objects.get(id=16635)
        EveMarketPrice.objects.create(eve_type=tungsten, average_price=7000)
        EveMarketPrice.objects.create(eve_type=mercury, average_price=9750)
        EveMarketPrice.objects.create(eve_type=evaporite_deposits, average_price=950)

    def test_should_return_value_per_unit(self):
        self.assertEqual(self.cinnebar.calc_refined_value_per_unit(0.7), 4002.25)


class TestEveOreTypeProfileUrl(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_return_correct_value(self):
        # given
        cinnebar = EveOreType.objects.get(id=45506)
        # when
        result = cinnebar.profile_url
        # then
        self.assertEqual(result, "https://www.kalkoken.org/apps/eveitems/?typeId=45506")


# class TestExtractionIsJackpot(NoSocketsTestCase):
#     @classmethod
#     def setUpClass(cls):
#         super().setUpClass()
#         load_eveuniverse()
#         load_allianceauth()
#         moon = helpers.create_moon_40161708()
#         owner = Owner.objects.create(
#             corporation=EveCorporationInfo.objects.get(corporation_id=2001)
#         )
#         cls.refinery = Refinery.objects.create(
#             id=40161708, moon=moon, owner=owner, eve_type_id=35835
#         )
#         cls.ore_quality_regular = EveOreType.objects.get(id=45490)
#         cls.ore_quality_improved = EveOreType.objects.get(id=46280)
#         cls.ore_quality_excellent = EveOreType.objects.get(id=46281)
#         cls.ore_quality_excellent_2 = EveOreType.objects.get(id=46283)

#     def test_should_be_jackpot(self):
#         # given
#         extraction = ExtractionFactory(
#             refinery=self.refinery,
#             chunk_arrival_at=now() + dt.timedelta(days=3),
#             auto_fracture_at=now() + dt.timedelta(days=4),
#             started_at=now() - dt.timedelta(days=3),
#             status=Extraction.Status.STARTED,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_excellent,
#             volume=1000000 * 0.1,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_excellent_2,
#             volume=1000000 * 0.1,
#         )
#         # when
#         result = extraction.calc_is_jackpot()
#         # then
#         self.assertTrue(result)

#     def test_should_not_be_jackpot_1(self):
#         # given
#         extraction = ExtractionFactory(
#             refinery=self.refinery,
#             chunk_arrival_at=now() + dt.timedelta(days=3),
#             auto_fracture_at=now() + dt.timedelta(days=4),
#             started_at=now() - dt.timedelta(days=3),
#             status=Extraction.Status.STARTED,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_excellent,
#             volume=1000000 * 0.1,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_improved,
#             volume=1000000 * 0.1,
#         )
#         # when
#         result = extraction.calc_is_jackpot()
#         # then
#         self.assertFalse(result)

#     def test_should_not_be_jackpot_2(self):
#         # given
#         extraction = ExtractionFactory(
#             refinery=self.refinery,
#             chunk_arrival_at=now() + dt.timedelta(days=3),
#             auto_fracture_at=now() + dt.timedelta(days=4),
#             started_at=now() - dt.timedelta(days=3),
#             status=Extraction.Status.STARTED,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_improved,
#             volume=1000000 * 0.1,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_excellent,
#             volume=1000000 * 0.1,
#         )
#         # when
#         result = extraction.calc_is_jackpot()
#         # then
#         self.assertFalse(result)

#     def test_should_not_be_jackpot_3(self):
#         # given
#         extraction = ExtractionFactory(
#             refinery=self.refinery,
#             chunk_arrival_at=now() + dt.timedelta(days=3),
#             auto_fracture_at=now() + dt.timedelta(days=4),
#             started_at=now() - dt.timedelta(days=3),
#             status=Extraction.Status.STARTED,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_regular,
#             volume=1000000 * 0.1,
#         )
#         ExtractionProductFactory(
#             extraction=extraction,
#             ore_type=self.ore_quality_improved,
#             volume=1000000 * 0.1,
#         )
#         # when
#         result = extraction.calc_is_jackpot()
#         # then
#         self.assertFalse(result)

#     def test_should_not_be_jackpot_4(self):
#         # given
#         extraction = ExtractionFactory(
#             refinery=self.refinery,
#             chunk_arrival_at=now() + dt.timedelta(days=3),
#             auto_fracture_at=now() + dt.timedelta(days=4),
#             started_at=now() - dt.timedelta(days=3),
#             status=Extraction.Status.STARTED,
#         )
#         # when
#         result = extraction.calc_is_jackpot()
#         # then
#         self.assertFalse(result)


class TestExtraction(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()

    def test_should_convert_to_calculated_extraction(self):
        # given
        refinery = RefineryFactory()
        my_map = [
            (Extraction.Status.STARTED, CalculatedExtraction.Status.STARTED),
            (Extraction.Status.CANCELED, CalculatedExtraction.Status.CANCELED),
            (Extraction.Status.READY, CalculatedExtraction.Status.READY),
            (Extraction.Status.COMPLETED, CalculatedExtraction.Status.COMPLETED),
        ]
        for in_status, out_status in my_map:
            with self.subTest(status=in_status):
                extraction = ExtractionFactory(status=in_status, refinery=refinery)
                # when
                obj = extraction.to_calculated_extraction()
                # then
                self.assertEqual(obj.status, out_status)


class TestOreQualityClass(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_return_correct_quality(self):
        # given
        ore_quality_regular = EveOreType.objects.get(id=EveTypeId.ZEOLITES)
        ore_quality_improved = EveOreType.objects.get(id=EveTypeId.BRIMFUL_ZEOLITES)
        ore_quality_excellent = EveOreType.objects.get(id=EveTypeId.GLISTENING_ZEOLITES)
        # when/then
        self.assertEqual(ore_quality_regular.quality_class, OreQualityClass.REGULAR)
        self.assertEqual(ore_quality_improved.quality_class, OreQualityClass.IMPROVED)
        self.assertEqual(ore_quality_excellent.quality_class, OreQualityClass.EXCELLENT)

    def test_should_return_correct_tag(self):
        self.assertIn("+100%", OreQualityClass.EXCELLENT.bootstrap_tag_html)



"""Notification models."""

from enum import Enum

from django.db import models
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveEntity

from app_utils.datetime import ldap_time_2_datetime

from moonmining.core import CalculatedExtraction, CalculatedExtractionProduct

# MAX_DISTANCE_TO_MOON_METERS = 3000000


class NotificationType(str, Enum):
    """ESI notification types used in this app."""

    MOONMINING_AUTOMATIC_FRACTURE = "MoonminingAutomaticFracture"
    MOONMINING_EXTRACTION_CANCELLED = "MoonminingExtractionCancelled"
    MOONMINING_EXTRACTION_FINISHED = "MoonminingExtractionFinished"
    MOONMINING_EXTRACTION_STARTED = "MoonminingExtractionStarted"
    MOONMINING_LASER_FIRED = "MoonminingLaserFired"

    def __str__(self) -> str:
        return self.value

    @classmethod
    def all_moon_mining(cls) -> set:
        """Return all moon mining notifications"""
        return {
            cls.MOONMINING_AUTOMATIC_FRACTURE,
            cls.MOONMINING_EXTRACTION_CANCELLED,
            cls.MOONMINING_EXTRACTION_FINISHED,
            cls.MOONMINING_EXTRACTION_STARTED,
            cls.MOONMINING_LASER_FIRED,
        }

    @classmethod
    def with_products(cls) -> set:
        """Return all notification types with have products."""
        return {
            cls.MOONMINING_AUTOMATIC_FRACTURE,
            cls.MOONMINING_EXTRACTION_FINISHED,
            cls.MOONMINING_EXTRACTION_STARTED,
            cls.MOONMINING_LASER_FIRED,
        }


class Notification(models.Model):
    """An EVE Online notification about structures."""

    # pk
    owner = models.ForeignKey(
        "Owner",
        on_delete=models.CASCADE,
        related_name="notifications",
        help_text=_("Corporation that received this notification"),
    )
    notification_id = models.PositiveBigIntegerField(verbose_name="id")
    # regular
    created = models.DateTimeField(
        null=True,
        default=None,
        help_text=_("Date when this notification was first received from ESI"),
    )
    details = models.JSONField(default=dict)
    notif_type = models.CharField(
        max_length=100,
        default="",
        db_index=True,
        verbose_name="type",
        help_text=_("type of this notification as reported by ESI"),
    )
    is_read = models.BooleanField(
        null=True,
        default=None,
        help_text=_("True when this notification has read in the eve client"),
    )
    last_updated = models.DateTimeField(
        help_text=_("Date when this notification has last been updated from ESI")
    )
    sender = models.ForeignKey(
        EveEntity, on_delete=models.CASCADE, null=True, default=None, related_name="+"
    )
    timestamp = models.DateTimeField(db_index=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["owner", "notification_id"], name="functional_pk_notification"
            )
        ]
        verbose_name = _("notification")
        verbose_name_plural = _("notifications")

    def __str__(self) -> str:
        return str(self.notification_id)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(notification_id={self.notification_id}, "
            f"owner='{self.owner}', notif_type='{self.notif_type}')"
        )

    def to_calculated_extraction(self) -> CalculatedExtraction:
        """Generate a calculated extraction from this notification."""
        params = {"refinery_id": self.details["structureID"]}
        if self.notif_type == NotificationType.MOONMINING_EXTRACTION_STARTED:
            params.update(
                {
                    "status": CalculatedExtraction.Status.STARTED,
                    "chunk_arrival_at": ldap_time_2_datetime(self.details["readyTime"]),
                    "auto_fracture_at": ldap_time_2_datetime(self.details["autoTime"]),
                    "started_at": self.timestamp,
                    "started_by": self.details.get("startedBy"),
                    "products": CalculatedExtractionProduct.create_list_from_dict(
                        self.details["oreVolumeByType"]
                    ),
                }
            )
        elif self.notif_type == NotificationType.MOONMINING_EXTRACTION_FINISHED:
            params.update(
                {
                    "status": CalculatedExtraction.Status.READY,
                    "auto_fracture_at": ldap_time_2_datetime(self.details["autoTime"]),
                    "products": CalculatedExtractionProduct.create_list_from_dict(
                        self.details["oreVolumeByType"]
                    ),
                }
            )
        elif self.notif_type in {
            NotificationType.MOONMINING_LASER_FIRED,
            NotificationType.MOONMINING_AUTOMATIC_FRACTURE,
        }:
            params.update(
                {
                    "fractured_by": self.details.get("firedBy"),
                    "fractured_at": self.timestamp,
                    "status": CalculatedExtraction.Status.COMPLETED,
                    "products": CalculatedExtractionProduct.create_list_from_dict(
                        self.details["oreVolumeByType"]
                    ),
                }
            )
        elif self.notif_type == NotificationType.MOONMINING_EXTRACTION_CANCELLED:
            params.update(
                {
                    "status": CalculatedExtraction.Status.CANCELED,
                    "canceled_at": self.timestamp,
                    "canceled_by": self.details.get("cancelledBy"),
                }
            )
        return CalculatedExtraction(**params)



"""Owner models."""

import datetime as dt
from collections import defaultdict
from typing import List, Optional, Tuple

import yaml

from django.contrib.auth.models import User
from django.db import models
from django.utils.html import format_html
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from esi.models import Token
from eveuniverse.models import EveEntity, EveMoon, EveSolarSystem, EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from app_utils.allianceauth import notify_admins_throttled
from app_utils.logging import LoggerAddTag
from app_utils.views import bootstrap_icon_plus_name_html

from moonmining import __title__
from moonmining.app_settings import MOONMINING_OVERWRITE_SURVEYS_WITH_ESTIMATES
from moonmining.constants import EveGroupId, EveTypeId, IconSize
from moonmining.core import CalculatedExtraction, CalculatedExtractionProduct
from moonmining.managers import MiningLedgerRecordManager, RefineryManager
from moonmining.models.extractions import EveOreType, Extraction
from moonmining.models.moons import Moon
from moonmining.models.notifications import Notification, NotificationType
from moonmining.providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class MiningLedgerRecord(models.Model):
    """A recorded mining activity in the vicinity of a refinery."""

    refinery = models.ForeignKey(
        "Refinery",
        on_delete=models.CASCADE,
        related_name="mining_ledger",
        help_text=_("Refinery this mining activity was observed at"),
    )
    day = models.DateField(db_index=True, help_text=_("last_updated in ESI"))
    character = models.ForeignKey(
        EveEntity,
        on_delete=models.CASCADE,
        related_name="+",
        help_text=_("character that did the mining"),
    )
    ore_type = models.ForeignKey(
        EveOreType, on_delete=models.CASCADE, related_name="mining_ledger"
    )
    # regular
    corporation = models.ForeignKey(
        EveEntity,
        on_delete=models.CASCADE,
        related_name="+",
        help_text=_("corporation of the character at time data was recorded"),
    )
    quantity = models.PositiveBigIntegerField()
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        related_name="mining_ledger",
    )

    objects = MiningLedgerRecordManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["refinery", "day", "character", "ore_type"],
                name="functional_pk_mining_activity",
            )
        ]
        verbose_name = _("ledger record")
        verbose_name_plural = _("ledger records")


class Owner(models.Model):
    """A EVE Online corporation owning refineries."""

    ESI_SERVICE_NAME_MOON_DRILLING = "Moon Drilling"

    # pk
    corporation = models.OneToOneField(
        EveCorporationInfo, on_delete=models.CASCADE, primary_key=True, related_name="+"
    )
    # regular
    character_ownership = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        related_name="+",
        help_text=_("Character used to sync this corporation from ESI"),
    )
    is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=_("Disabled corporations are excluded from the update process"),
    )
    last_update_at = models.DateTimeField(
        null=True, default=None, help_text=_("Time of last successful update")
    )
    last_update_ok = models.BooleanField(
        null=True, default=None, help_text=_("True if the last update was successful")
    )

    class Meta:
        verbose_name = _("owner")
        verbose_name_plural = _("owners")

    def __str__(self):
        return self.name

    @property
    def name(self) -> str:
        """Return name."""
        alliance_ticker_str = (
            f" [{self.corporation.alliance.alliance_ticker}]"
            if self.corporation.alliance
            else ""
        )
        return f"{self.corporation}{alliance_ticker_str}"

    @property
    def alliance_name(self) -> str:
        """Return alliance name."""
        return (
            self.corporation.alliance.alliance_name if self.corporation.alliance else ""
        )

    @property
    def name_html(self):
        """Return name as HTML."""
        return bootstrap_icon_plus_name_html(
            self.corporation.logo_url(size=IconSize.SMALL),
            self.name,
            size=IconSize.SMALL,
        )

    def fetch_token(self) -> Token:
        """Return valid token for this mining corp or raise exception on any error."""
        if not self.character_ownership:
            raise RuntimeError("This owner has no character configured.")
        token = (
            Token.objects.filter(
                character_id=self.character_ownership.character.character_id
            )
            .require_scopes(self.esi_scopes())
            .require_valid()
            .first()
        )
        if not token:
            raise Token.DoesNotExist(f"{self}: No valid token found.")
        return token

    def update_refineries_from_esi(self):
        """Update all refineries from ESI."""
        logger.info("%s: Updating refineries...", self)
        refineries = self._fetch_refineries_from_esi()
        for structure_id in refineries:
            try:
                self._update_or_create_refinery_from_esi(structure_id)
            except OSError as exc:
                exc_name = type(exc).__name__
                msg = (
                    f"{self}: Failed to fetch refinery with ID {structure_id} from ESI"
                )
                message_id = (
                    f"{__title__}-update_refineries_from_esi-"
                    f"{structure_id}-{exc_name}"
                )
                notify_admins_throttled(
                    message_id=message_id,
                    message=f"{msg}: {exc_name}: {exc}.",
                    title=f"{__title__}: Failed to fetch refinery",
                    level="warning",
                )
                logger.warning(msg, exc_info=True)
        # remove refineries that no longer exist
        self.refineries.exclude(id__in=refineries).delete()

        self.last_update_at = now()
        self.save()

    def _fetch_refineries_from_esi(self) -> dict:
        """Return current refineries with moon drills from ESI for this owner."""
        logger.info("%s: Fetching refineries from ESI...", self)
        structures = esi.client.Corporation.get_corporations_corporation_id_structures(
            corporation_id=self.corporation.corporation_id,
            token=self.fetch_token().valid_access_token(),
        ).results()
        refineries = {}
        for structure_info in structures:
            eve_type, _ = EveType.objects.get_or_create_esi(
                id=structure_info["type_id"]
            )
            structure_info["_eve_type"] = eve_type
            service_names = (
                {row["name"] for row in structure_info["services"]}
                if structure_info.get("services")
                else set()
            )
            if (
                eve_type.eve_group_id == EveGroupId.REFINERY
                and self.ESI_SERVICE_NAME_MOON_DRILLING in service_names
            ):
                refineries[structure_info["structure_id"]] = structure_info
        return refineries

    def _update_or_create_refinery_from_esi(self, structure_id: int):
        """Update or create a refinery with universe data from ESI."""
        logger.info("%s: Fetching details for refinery #%d", self, structure_id)
        structure_info = esi.client.Universe.get_universe_structures_structure_id(
            structure_id=structure_id, token=self.fetch_token().valid_access_token()
        ).results()
        refinery, _ = Refinery.objects.update_or_create(
            id=structure_id,
            defaults={
                "name": structure_info["name"],
                "eve_type": EveType.objects.get(id=structure_info["type_id"]),
                "owner": self,
            },
        )
        if not refinery.moon:
            refinery.update_moon_from_structure_info(structure_info)
        return True

    def fetch_notifications_from_esi(self) -> None:
        """fetches notification for the current owners and process them"""
        notifications = self._fetch_moon_notifications_from_esi()
        self._store_notifications(notifications)

    def _fetch_moon_notifications_from_esi(self) -> List[dict]:
        """Fetch all notifications from ESI for current owner."""
        logger.info("%s: Fetching notifications from ESI...", self)
        all_notifications = (
            esi.client.Character.get_characters_character_id_notifications(
                character_id=self.character_ownership.character.character_id,
                token=self.fetch_token().valid_access_token(),
            ).results()
        )
        moon_notifications = [
            notif
            for notif in all_notifications
            if notif["type"] in NotificationType.all_moon_mining()
        ]
        return moon_notifications

    def _store_notifications(self, notifications: list) -> int:
        """Store new notifications in database and return count of new objects."""
        # identify new notifications
        existing_notification_ids = set(
            self.notifications.values_list("notification_id", flat=True)
        )
        new_notifications = [
            obj
            for obj in notifications
            if obj["notification_id"] not in existing_notification_ids
        ]
        # create new notif objects
        sender_type_map = {
            "character": EveEntity.CATEGORY_CHARACTER,
            "corporation": EveEntity.CATEGORY_CORPORATION,
            "alliance": EveEntity.CATEGORY_ALLIANCE,
        }
        new_notification_objects = []
        for notification in new_notifications:
            known_sender_type = sender_type_map.get(notification["sender_type"])
            if known_sender_type:
                sender, _ = EveEntity.objects.get_or_create_esi(
                    id=notification["sender_id"]
                )
            else:
                sender = None
            text = notification["text"] if "text" in notification else None
            is_read = notification["is_read"] if "is_read" in notification else None
            new_notification_objects.append(
                Notification(
                    notification_id=notification["notification_id"],
                    owner=self,
                    created=now(),
                    details=yaml.safe_load(text) if text else {},
                    is_read=is_read,
                    last_updated=now(),
                    # at least one type has a trailing white space
                    # which we need to remove
                    notif_type=notification["type"].strip(),
                    sender=sender,
                    timestamp=notification["timestamp"],
                )
            )

        Notification.objects.bulk_create(new_notification_objects)
        if len(new_notification_objects) > 0:
            logger.info(
                "%s: Received %d new notifications from ESI",
                self,
                len(new_notification_objects),
            )
        else:
            logger.info("%s: No new notifications received from ESI", self)
        return len(new_notification_objects)

    def update_extractions(self):
        """Update extractions fro ESI."""
        self.update_extractions_from_esi()
        Extraction.objects.all().update_status()
        self.update_extractions_from_notifications()

    def update_extractions_from_esi(self):
        """Creates new extractions from ESI for current owner."""
        extractions_by_refinery = self._fetch_extractions_from_esi()
        self._update_or_create_extractions(extractions_by_refinery)

    def _fetch_extractions_from_esi(self):
        logger.info("%s: Fetching extractions from ESI...", self)
        extractions = (
            esi.client.Industry.get_corporation_corporation_id_mining_extractions(
                corporation_id=self.corporation.corporation_id,
                token=self.fetch_token().valid_access_token(),
            ).results()
        )
        logger.info("%s: Received %d extractions from ESI.", self, len(extractions))
        extractions_by_refinery = defaultdict(list)
        for row in extractions:
            extractions_by_refinery[row["structure_id"]].append(row)
        return extractions_by_refinery

    def _update_or_create_extractions(self, extractions_by_refinery: dict) -> None:
        new_extractions_count = 0
        for refinery_id, refinery_extractions in extractions_by_refinery.items():
            try:
                refinery = self.refineries.get(pk=refinery_id)
            except Refinery.DoesNotExist:
                continue
            new_extractions_count += refinery.create_extractions_from_esi_response(
                refinery_extractions
            )
            refinery.cancel_started_extractions_missing_from_list(
                [row["extraction_start_time"] for row in refinery_extractions]
            )
        if new_extractions_count:
            logger.info("%s: Created %d new extractions.", self, new_extractions_count)

    def update_extractions_from_notifications(self):
        """Create or update extractions from notifications."""
        logger.info("%s: Updating extractions from notifications...", self)
        notifications_count = self.notifications.count()
        if not notifications_count:
            logger.info("%s: No moon notifications.", self)
            return

        logger.info("%s: Processing %d moon notifications.", self, notifications_count)
        for refinery in self.refineries.all():
            _update_extractions_for_refinery(self, refinery)

    def fetch_mining_ledger_observers_from_esi(self) -> set:
        """Fetch mining ledger observers from ESI and return them."""
        logger.info("%s: Fetching mining observers from ESI...", self)
        observers = esi.client.Industry.get_corporation_corporation_id_mining_observers(
            corporation_id=self.corporation.corporation_id,
            token=self.fetch_token().valid_access_token(),
        ).results()
        logger.info("%s: Received %d observers from ESI.", self, len(observers))
        return {
            row["observer_id"]
            for row in observers
            if row["observer_type"] == "structure"
        }

    @classmethod
    def esi_scopes(cls):
        """Return list of all required esi scopes."""
        return [
            "esi-industry.read_corporation_mining.v1",
            "esi-universe.read_structures.v1",
            "esi-characters.read_notifications.v1",
            "esi-corporations.read_structures.v1",
            "esi-industry.read_corporation_mining.v1",
        ]


class Refinery(models.Model):
    """An Eve Online refinery structure."""

    # pk
    id = models.PositiveBigIntegerField(primary_key=True)
    # regular
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")
    moon = models.OneToOneField(
        Moon,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        related_name="refinery",
        help_text=_("The moon this refinery is anchored at (if any)"),
    )
    name = models.CharField(max_length=150, db_index=True)
    owner = models.ForeignKey(
        Owner,
        on_delete=models.CASCADE,
        related_name="refineries",
        help_text=_("Corporation that owns this refinery"),
    )
    ledger_last_update_at = models.DateTimeField(
        null=True, default=None, help_text=_("last successful update of mining ledger")
    )
    ledger_last_update_ok = models.BooleanField(
        null=True,
        default=None,
        help_text=_("True if the last update of the mining ledger was successful"),
    )

    objects = RefineryManager()

    class Meta:
        verbose_name = _("refinery")
        verbose_name_plural = _("refineries")

    def __str__(self):
        return self.name

    def name_html(self) -> str:
        """Return name as HTML."""
        return format_html("{}<br>{}", self.name, self.owner.name)

    def update_moon_from_structure_info(self, structure_info: dict) -> bool:
        """Find moon based on location in space and update the object.
        Returns True when successful, else false
        """
        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=structure_info["solar_system_id"]
        )
        try:
            nearest_celestial = solar_system.nearest_celestial(
                x=structure_info["position"]["x"],
                y=structure_info["position"]["y"],
                z=structure_info["position"]["z"],
                group_id=EveGroupId.MOON,
            )
        except OSError:
            logger.exception("%s: Failed to fetch nearest celestial ", self)
            return False
        if not nearest_celestial or nearest_celestial.eve_type.id != EveTypeId.MOON:
            return False
        eve_moon = nearest_celestial.eve_object
        moon, _ = Moon.objects.get_or_create(eve_moon=eve_moon)
        self.moon = moon
        self.save()
        return True

    def update_moon_from_eve_id(self, eve_moon_id: int):
        """Update moon from ESI."""
        eve_moon, _ = EveMoon.objects.get_or_create_esi(id=eve_moon_id)
        moon, _ = Moon.objects.get_or_create(eve_moon=eve_moon)
        self.moon = moon
        self.save()

    def update_mining_ledger_from_esi(self):
        """Update mining ledger from ESI."""
        self._reset_update_status()
        records = self._fetch_ledger_from_esi()
        self._preload_missing_ore_types(records)
        self._store_ledger(records)
        self._record_successful_update()

    def _reset_update_status(self):
        self.ledger_last_update_at = now()
        self.ledger_last_update_ok = None
        self.save()

    def _fetch_ledger_from_esi(self):
        logger.debug("%s: Fetching mining observer records from ESI...", self)
        token = self.owner.fetch_token().valid_access_token()
        records = esi.client.Industry.get_corporation_corporation_id_mining_observers_observer_id(
            corporation_id=self.owner.corporation.corporation_id,
            observer_id=self.id,
            token=token,
        ).results()
        logger.info(
            "%s: Received %d mining observer records from ESI", self, len(records)
        )

        return records

    def _preload_missing_ore_types(self, records):
        EveOreType.objects.bulk_get_or_create_esi(
            ids=[record["type_id"] for record in records]
        )

    def _store_ledger(self, records):
        character_2_user = {
            obj[0]: obj[1]
            for obj in CharacterOwnership.objects.values_list(
                "character__character_id",
                "user_id",
            )
        }
        entity_ids = set()
        for record in records:
            character, _ = EveEntity.objects.get_or_create(id=record["character_id"])
            corporation, _ = EveEntity.objects.get_or_create(
                id=record["recorded_corporation_id"]
            )
            entity_ids.add(character.id)
            entity_ids.add(corporation.id)
            MiningLedgerRecord.objects.update_or_create(
                refinery=self,
                character=character,
                day=record["last_updated"],
                ore_type_id=record["type_id"],
                defaults={
                    "corporation": corporation,
                    "quantity": record["quantity"],
                    "user_id": character_2_user.get(character.id),
                },
            )

        try:
            EveEntity.objects.bulk_resolve_ids(entity_ids)
        except OSError:
            logger.warning(
                "%s: Failed to resolve entity IDs for mining ledger: %s",
                self,
                entity_ids,
                exc_info=True,
            )

    def _record_successful_update(self):
        self.ledger_last_update_ok = True
        self.save()

    def create_extractions_from_esi_response(self, esi_extractions: List[dict]) -> int:
        """Create extractions from an ESI repose and return number of created objs."""
        existing_extractions = set(
            self.extractions.values_list("started_at", flat=True)
        )
        new_extractions = []
        for esi_extraction in esi_extractions:
            extraction_start_time = esi_extraction["extraction_start_time"]
            if extraction_start_time not in existing_extractions:
                chunk_arrival_time = esi_extraction["chunk_arrival_time"]
                auto_fracture_at = esi_extraction["natural_decay_time"]
                if now() > auto_fracture_at:
                    status = Extraction.Status.COMPLETED
                elif now() > chunk_arrival_time:
                    status = Extraction.Status.READY
                else:
                    status = Extraction.Status.STARTED
                new_extractions.append(
                    Extraction(
                        refinery=self,
                        chunk_arrival_at=esi_extraction["chunk_arrival_time"],
                        started_at=extraction_start_time,
                        status=status,
                        auto_fracture_at=auto_fracture_at,
                    )
                )
        if new_extractions:
            Extraction.objects.bulk_create(new_extractions, batch_size=500)
        return len(new_extractions)

    def cancel_started_extractions_missing_from_list(
        self, started_at_list: List[dt.datetime]
    ) -> int:
        """Cancel started extractions that are not included in given list."""
        canceled_extractions_qs = self.extractions.filter(
            status=Extraction.Status.STARTED
        ).exclude(started_at__in=started_at_list)
        canceled_extractions_count = canceled_extractions_qs.count()
        if canceled_extractions_count:
            logger.info(
                "%s: Found %d likely canceled extractions.",
                self,
                canceled_extractions_count,
            )
            canceled_extractions_qs.update(
                status=Extraction.Status.CANCELED, canceled_at=now()
            )
        return canceled_extractions_count


def _update_extractions_for_refinery(owner: Owner, refinery: Refinery):
    notifications_for_refinery = owner.notifications.filter(
        details__structureID=refinery.id
    )
    if not refinery.moon and notifications_for_refinery.exists():
        # Update the refinery's moon from notification in case
        # it was not found by nearest_celestial.
        notif = notifications_for_refinery.first()
        refinery.update_moon_from_eve_id(notif.details["moonID"])

    extraction, updated_count = _find_extraction_for_refinery(
        refinery, notifications_for_refinery
    )
    if extraction:
        updated = Extraction.objects.update_from_calculated(extraction)
        updated_count += 1 if updated else 0

    if updated_count:
        logger.info(
            "%s: %s: Updated %d extractions from notifications",
            owner,
            refinery,
            updated_count,
        )


def _find_extraction_for_refinery(
    refinery: Refinery,
    notifications_for_refinery: models.QuerySet["Notification"],
) -> Tuple[Optional[CalculatedExtraction], int]:
    extraction: Optional[CalculatedExtraction] = None
    updated_count = 0
    for notif in notifications_for_refinery.order_by("timestamp"):
        if notif.notif_type == NotificationType.MOONMINING_EXTRACTION_STARTED:
            extraction = notif.to_calculated_extraction()
            if refinery.moon.update_products_from_calculated_extraction(
                extraction,
                overwrite_survey=MOONMINING_OVERWRITE_SURVEYS_WITH_ESTIMATES,
            ):
                logger.info("%s: Products updated from extraction", refinery.moon)

        elif extraction:
            if extraction.status == CalculatedExtraction.Status.STARTED:
                if notif.notif_type == NotificationType.MOONMINING_EXTRACTION_CANCELLED:
                    extraction.status = CalculatedExtraction.Status.CANCELED
                    extraction.canceled_at = notif.timestamp
                    extraction.canceled_by = notif.details.get("cancelledBy")
                    updated = Extraction.objects.update_from_calculated(extraction)
                    updated_count += 1 if updated else 0
                    extraction = None

                elif (
                    notif.notif_type == NotificationType.MOONMINING_EXTRACTION_FINISHED
                ):
                    extraction.status = CalculatedExtraction.Status.READY
                    extraction.products = (
                        CalculatedExtractionProduct.create_list_from_dict(
                            notif.details["oreVolumeByType"]
                        )
                    )

            elif extraction.status == CalculatedExtraction.Status.READY:
                if notif.notif_type == NotificationType.MOONMINING_LASER_FIRED:
                    extraction.status = CalculatedExtraction.Status.COMPLETED
                    extraction.fractured_at = notif.timestamp
                    extraction.fractured_by = notif.details.get("firedBy")
                    extraction.products = (
                        CalculatedExtractionProduct.create_list_from_dict(
                            notif.details["oreVolumeByType"]
                        )
                    )
                    updated = Extraction.objects.update_from_calculated(extraction)
                    updated_count += 1 if updated else 0
                    extraction = None

                elif notif.notif_type == NotificationType.MOONMINING_AUTOMATIC_FRACTURE:
                    extraction.status = CalculatedExtraction.Status.COMPLETED
                    extraction.fractured_at = notif.timestamp
                    extraction.products = (
                        CalculatedExtractionProduct.create_list_from_dict(
                            notif.details["oreVolumeByType"]
                        )
                    )
                    updated = Extraction.objects.update_from_calculated(extraction)
                    updated_count += 1 if updated else 0
                    extraction = None
        else:
            if notif.notif_type == NotificationType.MOONMINING_EXTRACTION_FINISHED:
                extraction = notif.to_calculated_extraction()

    return extraction, updated_count



"""Alliance Auth app for tracking moon extractions and scouting new moons."""

# pylint: disable = invalid-name
default_app_config = "moonmining.apps.MoonMiningConfig"

__version__ = "2.1.0"
__title__ = "Moon Mining"



"""Moon models."""

from typing import List, Optional

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models, transaction
from django.db.models import F, Sum, Value
from django.db.models.functions import Coalesce
from django.utils.html import format_html
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveMoon, EveRegion, EveSolarSystem

from app_utils.views import BootstrapStyleBS5

from moonmining.app_settings import (
    MOONMINING_VOLUME_PER_DAY,
    MOONMINING_VOLUME_PER_MONTH,
)
from moonmining.core import CalculatedExtraction
from moonmining.helpers import bootstrap5_label_html
from moonmining.managers import MoonManager
from moonmining.models.extractions import EveOreType, OreRarityClass


class Label(models.Model):
    """A custom label for structuring moons."""

    class Style(models.TextChoices):
        """A label style."""

        DARK_BLUE = "primary", _("dark blue")
        GREEN = "success", _("green")
        GREY = "default", _("grey")
        LIGHT_BLUE = "info", _("light blue")
        ORANGE = "warning", _("orange")
        RED = "danger", _("red")

        @property
        def bootstrap_style(self) -> str:
            """Return HTML to render a bootstrap tag."""
            map_to_type = {
                self.DARK_BLUE: BootstrapStyleBS5.PRIMARY,
                self.GREEN: BootstrapStyleBS5.SUCCESS,
                self.LIGHT_BLUE: BootstrapStyleBS5.INFO,
                self.ORANGE: BootstrapStyleBS5.WARNING,
                self.RED: BootstrapStyleBS5.DANGER,
            }
            try:
                return map_to_type[self].value
            except KeyError:
                return BootstrapStyleBS5.DEFAULT

    description = models.TextField(default="", blank=True)
    name = models.CharField(max_length=100, unique=True)
    style = models.CharField(max_length=16, choices=Style.choices, default=Style.GREY)

    class Meta:
        verbose_name = _("label")
        verbose_name_plural = _("labels")

    def __str__(self) -> str:
        return self.name

    @property
    def tag_html(self) -> str:
        """Return tag HTML for this obj."""
        label_style = self.Style(self.style).bootstrap_style
        return bootstrap5_label_html(self.name, label=label_style)


class Moon(models.Model):
    """Known moon through either survey data or anchored refinery.

    "Head" model for many of the other models.
    """

    # pk
    eve_moon = models.OneToOneField(
        EveMoon, on_delete=models.CASCADE, primary_key=True, related_name="+"
    )
    # regular
    label = models.ForeignKey(
        Label, on_delete=models.SET_DEFAULT, default=None, null=True
    )
    products_updated_at = models.DateTimeField(
        null=True, default=None, help_text=_("Time the last moon survey was uploaded")
    )
    products_updated_by = models.ForeignKey(
        User,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        help_text=_("User who uploaded the last moon survey"),
    )
    rarity_class = models.PositiveIntegerField(
        choices=OreRarityClass.choices, default=OreRarityClass.NONE
    )
    value = models.FloatField(
        null=True,
        default=None,
        validators=[MinValueValidator(0.0)],
        db_index=True,
        help_text=_("Calculated value estimate"),
    )

    objects = MoonManager()

    class Meta:
        verbose_name = _("moon")
        verbose_name_plural = _("moons")

    def __str__(self):
        return self.name

    @property
    def name(self) -> str:
        """Return name of this moon."""
        return self.eve_moon.name.replace("Moon ", "")

    def region(self) -> EveRegion:
        """Return region."""
        return self.solar_system().eve_constellation.eve_region

    def solar_system(self) -> EveSolarSystem:
        """Return solar system."""
        return self.eve_moon.eve_planet.eve_solar_system

    @property
    def is_owned(self) -> bool:
        """Return True when this moon has a known owner, else False."""
        return hasattr(self, "refinery")

    @property
    def rarity_tag_html(self) -> str:
        """Return rarity tag HTML for this moon."""
        return OreRarityClass(self.rarity_class).bootstrap_tag_html

    def labels_html(self) -> str:
        """Generate HTML with all labels."""
        tags = [self.rarity_tag_html]
        if self.label:
            tags.append(self.label.tag_html)
        return format_html(" ".join(tags))

    def products_sorted(self) -> models.QuerySet:
        """Return current products."""
        try:
            return (
                self.products.select_related(
                    "ore_type", "ore_type__eve_group", "ore_type__extras"
                )
                .annotate(total_price=self._total_price_db_func())
                .order_by("ore_type__name")
            )
        except (ObjectDoesNotExist, AttributeError):
            return type(self).objects.none()

    def calc_rarity_class(self) -> Optional[OreRarityClass]:
        """Return rarity class of this moon."""
        try:
            return max(
                OreRarityClass.from_eve_group_id(eve_group_id)
                for eve_group_id in self.products.select_related(
                    "ore_type"
                ).values_list("ore_type__eve_group_id", flat=True)
            )
        except (ObjectDoesNotExist, ValueError):
            return OreRarityClass.NONE

    def calc_value(self) -> Optional[float]:
        """Calculate value estimate."""
        try:
            return self.products.aggregate(total_value=self._total_price_db_func())[
                "total_value"
            ]
        except (ObjectDoesNotExist, KeyError, AttributeError):
            return None

    @staticmethod
    def _total_price_db_func():
        return Sum(
            Coalesce(F("ore_type__extras__current_price"), 0.0)
            * F("amount")
            * Value(float(MOONMINING_VOLUME_PER_MONTH))
            / F("ore_type__volume"),
            output_field=models.FloatField(),
        )

    def update_calculated_properties(self):
        """Update all calculated properties for this moon."""
        self.value = self.calc_value()
        self.rarity_class = self.calc_rarity_class()
        self.save()

    def update_products(
        self, moon_products: List["MoonProduct"], updated_by: Optional[User] = None
    ) -> None:
        """Update products of this moon."""
        with transaction.atomic():
            self.products.all().delete()
            MoonProduct.objects.bulk_create(moon_products, batch_size=500)
        self.products_updated_at = now()
        self.products_updated_by = updated_by
        self.update_calculated_properties()

    def update_products_from_calculated_extraction(
        self, extraction: CalculatedExtraction, overwrite_survey: bool = False
    ) -> bool:
        """Replace moon product with calculated values from this extraction.

        Returns True if update was done, else False
        """
        if extraction.products and (
            overwrite_survey or self.products_updated_by is None
        ):
            moon_products = [
                MoonProduct(
                    moon=self,
                    amount=product.amount,
                    ore_type=EveOreType.objects.get_or_create_esi(
                        id=product.ore_type_id
                    )[0],
                )
                for product in extraction.moon_products_estimated(
                    MOONMINING_VOLUME_PER_DAY
                )
            ]
            self.update_products(moon_products)
            return True
        return False

    def update_products_from_latest_extraction(
        self, overwrite_survey: bool = False
    ) -> Optional[bool]:
        """Update products from latest extractions and return if successful."""
        try:
            extraction = self.refinery.extractions.order_by("-started_at").first()
        except ObjectDoesNotExist:
            return None

        if not extraction:
            return None

        success = self.update_products_from_calculated_extraction(
            extraction.to_calculated_extraction(), overwrite_survey=overwrite_survey
        )
        return success


class MoonProduct(models.Model):
    """A product of a moon, i.e. a specific ore."""

    moon = models.ForeignKey(Moon, on_delete=models.CASCADE, related_name="products")
    ore_type = models.ForeignKey(EveOreType, on_delete=models.CASCADE, related_name="+")

    amount = models.FloatField(
        validators=[MinValueValidator(0.0), MaxValueValidator(1.0)]
    )

    def __str__(self):
        return f"{self.ore_type.name} - {self.amount}"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["moon", "ore_type"], name="functional_pk_moonproduct"
            )
        ]
        verbose_name = _("moon product")
        verbose_name_plural = _("moons products")

    @property
    def amount_percent(self) -> float:
        """Return the amount of this product as percent"""
        return self.amount * 100



"""Helpers for views."""

from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from app_utils.views import BootstrapStyleBS5

from moonmining.models import Moon


def moon_details_button_html(moon: Moon) -> str:
    """Return HTML to render a moon details button."""
    return fontawesome_modal_button_html(
        modal_id="modalMoonDetails",
        fa_code="fas fa-moon",
        ajax_url=reverse("moonmining:moon_details", args=[moon.pk]),
        tooltip=_("Moon details"),
    )


def fontawesome_modal_button_html(
    modal_id: str,
    fa_code: str,
    ajax_url: str = "",
    tooltip: str = "",
    style=BootstrapStyleBS5.DEFAULT,
) -> str:
    """Return HTML for a modal button with fontawesome symbols.

    Args:
        modal_id: DOM ID of modal to invoke
        fa_code: fontawesome code, e.g. "fas fa-moon"
        ajax_url: URL to invoke via AJAX for loading modal content
        tooltip: text to appear as tooltip
        style: Bootstrap context style for the button
    """
    return format_html(
        '<button type="button" '
        'class="btn btn-{}" '
        'data-bs-toggle="modal" '
        'data-bs-target="#{}" '
        "{}"
        "{}>"
        '<i class="{}"></i>'
        "</button>",
        BootstrapStyleBS5(style),
        modal_id,
        mark_safe(f'title="{tooltip}" ') if tooltip else "",
        mark_safe(f'data-ajax_url="{ajax_url}" ') if ajax_url else "",
        fa_code,
    )



# flake8: noqa

from .extractions import (
    EveOreType,
    EveOreTypeExtras,
    Extraction,
    ExtractionProduct,
    OreQualityClass,
    OreRarityClass,
)
from .general import General
from .moons import Label, Moon, MoonProduct
from .notifications import Notification, NotificationType
from .owners import MiningLedgerRecord, Owner, Refinery



"""App settings."""

from app_utils.app_settings import clean_setting

MOONMINING_COMPLETED_EXTRACTIONS_HOURS_UNTIL_STALE = clean_setting(
    "MOONMINING_COMPLETED_EXTRACTIONS_HOURS_UNTIL_STALE", 12
)
"""Number of hours an extractions that has passed its ready time is still shown
on the upcoming extractions tab.
"""

MOONMINING_REPROCESSING_YIELD = clean_setting("MOONMINING_REPROCESSING_YIELD", 0.85)
"""Reprocessing yield used for calculating reprocess prices."""

MOONMINING_USE_REPROCESS_PRICING = clean_setting(
    "MOONMINING_USE_REPROCESS_PRICING", False
)
"""Whether to calculate prices from reprocessed materials or not.
Will use direct ore prices when switched off.
"""

MOONMINING_VOLUME_PER_DAY = clean_setting("MOONMINING_VOLUME_PER_DAY", 960_400)
MOONMINING_DAYS_PER_MONTH = clean_setting("MOONMINING_DAYS_PER_MONTH", 30.4)
MOONMINING_VOLUME_PER_MONTH = MOONMINING_VOLUME_PER_DAY * MOONMINING_DAYS_PER_MONTH
"""Total ore volume per month used for calculating moon values."""

MOONMINING_ADMIN_NOTIFICATIONS_ENABLED = clean_setting(
    "MOONMINING_ADMIN_NOTIFICATIONS_ENABLED", True
)
"""Whether admins will get notifications about important events like
when someone adds a new owner.
"""

MOONMINING_OVERWRITE_SURVEYS_WITH_ESTIMATES = clean_setting(
    "MOONMINING_OVERWRITE_SURVEYS_WITH_ESTIMATES", False
)
"""whether uploaded survey are automatically overwritten by product estimates from
extractions to keep the moon values current."""



from django.test import RequestFactory, TestCase
from eveuniverse.models import EveMoon

from app_utils.testing import create_user_from_evecharacter

from moonmining.models import Label, Moon, Owner
from moonmining.tests import helpers
from moonmining.tests.testdata.factories import MoonFactory, RefineryFactory
from moonmining.tests.testdata.load_allianceauth import load_allianceauth
from moonmining.tests.testdata.load_eveuniverse import load_eveuniverse
from moonmining.views import moons

MODULE_PATH = "moonmining.views.moons"


class TestMoonsData(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_market_prices()
        cls.moon = MoonFactory(eve_moon=EveMoon.objects.get(id=40161708))
        cls.moon.label = Label.objects.create(name="Dummy")
        cls.moon.save()
        MoonFactory(eve_moon=EveMoon.objects.get(id=40131695))
        MoonFactory(eve_moon=EveMoon.objects.get(id=40161709))

    @staticmethod
    def _response_to_dict(response):
        data = helpers.json_response_to_python_2(response)
        return {int(obj[0]): obj for obj in data}

    def test_should_return_all_moons(self):
        # given
        user, _ = create_user_from_evecharacter(
            1001,
            permissions=["moonmining.basic_access", "moonmining.view_all_moons"],
            scopes=Owner.esi_scopes(),
        )
        request = self.factory.get("/")
        request.user = user
        my_view = moons.MoonListJson.as_view()

        # when
        response = my_view(request, category=moons.MoonsCategory.ALL)

        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {40131695, 40161708, 40161709})
        obj = data[40161708]
        self.assertEqual(obj[1], "Auga V - 1")

    def test_should_return_our_moons_only(self):
        # given
        moon = Moon.objects.get(pk=40131695)
        RefineryFactory(moon=moon)
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=["moonmining.basic_access", "moonmining.extractions_access"],
            scopes=Owner.esi_scopes(),
        )
        request = self.factory.get("/")
        request.user = user
        my_view = moons.MoonListJson.as_view()

        # when
        response = my_view(request, category=moons.MoonsCategory.OURS)

        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {40131695})

    def test_should_return_our_moons_when_all_moons_perm(self):
        # given
        moon = Moon.objects.get(pk=40131695)
        RefineryFactory(moon=moon)
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=["moonmining.basic_access", "moonmining.extractions_access"],
            scopes=Owner.esi_scopes(),
        )
        request = self.factory.get("/")
        request.user = user
        my_view = moons.MoonListJson.as_view()

        # when
        response = my_view(request, category=moons.MoonsCategory.OURS)

        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {40131695})

    def test_should_handle_empty_refineries(self):
        # given
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=["moonmining.basic_access", "moonmining.extractions_access"],
            scopes=Owner.esi_scopes(),
        )
        moon = Moon.objects.get(pk=40131695)
        refinery = RefineryFactory(moon=moon)
        RefineryFactory(owner=refinery.owner, moon=None)
        request = self.factory.get("/")
        request.user = user
        my_view = moons.MoonListJson.as_view()
        # when
        response = my_view(request, category=moons.MoonsCategory.OURS)
        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {40131695})

    def test_should_return_uploaded_moons_only(self):
        # given
        user, _ = create_user_from_evecharacter(
            1001,
            permissions=["moonmining.basic_access", "moonmining.upload_moon_scan"],
            scopes=Owner.esi_scopes(),
        )
        self.moon.products_updated_by = user
        self.moon.save()
        request = self.factory.get("/")
        request.user = user
        my_view = moons.MoonListJson.as_view()
        # when
        response = my_view(request, category=moons.MoonsCategory.UPLOADS)
        # then
        self.assertEqual(response.status_code, 200)
        data = self._response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {40161708})


class TestMoonInfo(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()

    def test_should_open_page(self):
        # given
        moon = MoonFactory()
        user, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.view_all_moons",
                "moonmining.upload_moon_scan",
                "moonmining.add_refinery_owner",
            ],
            scopes=[
                "esi-industry.read_corporation_mining.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-corporations.read_structures.v1",
            ],
        )
        self.client.force_login(user)
        # when
        response = self.client.get(f"/moonmining/moon/{moon.pk}")
        # then
        self.assertTemplateUsed(response, "moonmining/modals/moon_details.html")



from unittest.mock import Mock, patch

from django.contrib.sessions.middleware import SessionMiddleware
from django.http import Http404
from django.test import RequestFactory, TestCase
from django.urls import reverse
from esi.models import Token

from allianceauth.eveonline.models import EveCorporationInfo
from app_utils.testing import create_user_from_evecharacter

from moonmining.models import Owner
from moonmining.tests.testdata.factories import (
    ExtractionFactory,
    MoonFactory,
    RefineryFactory,
)
from moonmining.tests.testdata.load_allianceauth import load_allianceauth
from moonmining.tests.testdata.load_eveuniverse import load_eveuniverse
from moonmining.views import general

MODULE_PATH = "moonmining.views.general"


class TestOwner(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001, permissions=["moonmining.add_refinery_owner"]
        )

    @patch(MODULE_PATH + ".notify_admins")
    @patch(MODULE_PATH + ".tasks.update_owner")
    @patch(MODULE_PATH + ".messages")
    def test_should_add_new_owner(
        self, mock_messages, mock_update_owner, mock_notify_admins
    ):
        # given
        token = Mock(spec=Token)
        token.character_id = self.character_ownership.character.character_id
        request = self.factory.get(reverse("moonmining:add_owner"))
        request.user = self.user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = general.add_owner.__wrapped__.__wrapped__.__wrapped__
        # when
        response = orig_view(request, token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("moonmining:index"))
        self.assertTrue(mock_messages.success.called)
        self.assertTrue(mock_update_owner.delay.called)
        self.assertTrue(mock_notify_admins.called)
        obj = Owner.objects.get(corporation__corporation_id=2001)
        self.assertEqual(obj.character_ownership, self.character_ownership)

    @patch(MODULE_PATH + ".tasks.update_owner")
    @patch(MODULE_PATH + ".messages")
    def test_should_update_existing_owner(self, mock_messages, mock_update_owner):
        # given
        Owner.objects.create(
            corporation=EveCorporationInfo.objects.get(corporation_id=2001),
            character_ownership=None,
        )
        token = Mock(spec=Token)
        token.character_id = self.character_ownership.character.character_id
        request = self.factory.get(reverse("moonmining:add_owner"))
        request.user = self.user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = general.add_owner.__wrapped__.__wrapped__.__wrapped__
        # when
        response = orig_view(request, token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("moonmining:index"))
        self.assertTrue(mock_messages.success.called)
        self.assertTrue(mock_update_owner.delay.called)
        obj = Owner.objects.get(corporation__corporation_id=2001)
        self.assertEqual(obj.character_ownership, self.character_ownership)

    @patch(MODULE_PATH + ".tasks.update_owner")
    @patch(MODULE_PATH + ".messages")
    def test_should_raise_404_if_character_ownership_not_found(
        self, mock_messages, mock_update_owner
    ):
        # given
        token = Mock(spec=Token)
        token.character_id = 1099
        request = self.factory.get(reverse("moonmining:add_owner"))
        request.user = self.user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = general.add_owner.__wrapped__.__wrapped__.__wrapped__
        # when
        with self.assertRaises(Http404):
            orig_view(request, token)


class TestViewsAreWorking(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        cls.user, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.reports_access",
                "moonmining.view_all_moons",
                "moonmining.upload_moon_scan",
                "moonmining.add_refinery_owner",
            ],
            scopes=[
                "esi-industry.read_corporation_mining.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-corporations.read_structures.v1",
            ],
        )
        cls.moon = MoonFactory()

    def test_should_redirect_to_extractions_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/")
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/moonmining/extractions")

    def test_should_open_extractions_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/extractions")
        # then
        self.assertTemplateUsed(response, "moonmining/extractions.html")

    def test_should_open_moon_details_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get(f"/moonmining/moon/{self.moon.pk}?new_page=yes")
        # then
        self.assertTemplateUsed(response, "moonmining/_generic_modal_page.html")

    def test_should_open_extraction_details_page(self):
        # given
        refinery = RefineryFactory(moon=self.moon)
        extraction = ExtractionFactory(refinery=refinery)
        self.client.force_login(self.user)
        # when
        response = self.client.get(
            f"/moonmining/extraction/{extraction.pk}?new_page=yes"
        )
        # then
        self.assertTemplateUsed(response, "moonmining/_generic_modal_page.html")

    def test_should_open_add_moon_scan_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/upload_survey")
        # then
        self.assertTemplateUsed(response, "moonmining/modals/upload_survey.html")

    def test_should_open_moons_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/moons")
        # then
        self.assertTemplateUsed(response, "moonmining/moons.html")

    def test_should_open_reports_page(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/reports")
        # then
        self.assertTemplateUsed(response, "moonmining/reports.html")

    def test_should_handle_empty_refineries_extractions_page(self):
        # given
        refinery = RefineryFactory()
        ExtractionFactory(refinery=refinery)
        self.client.force_login(self.user)
        # when
        response = self.client.get("/moonmining/extractions")
        # then
        self.assertTemplateUsed(response, "moonmining/extractions.html")



import datetime as dt
from unittest.mock import patch

from django.utils.timezone import now
from eveuniverse.models import EveMarketPrice

from app_utils.testdata_factories import UserFactory
from app_utils.testing import NoSocketsTestCase

from moonmining.constants import EveTypeId
from moonmining.core import CalculatedExtractionProduct
from moonmining.models import EveOreType, OreRarityClass
from moonmining.tests import helpers
from moonmining.tests.testdata.factories import (
    CalculatedExtractionFactory,
    ExtractionFactory,
    MoonFactory,
    MoonProductFactory,
    RefineryFactory,
)
from moonmining.tests.testdata.load_allianceauth import load_allianceauth
from moonmining.tests.testdata.load_eveuniverse import load_eveuniverse

MODELS_PATH = "moonmining.models"


class TestMoonUpdateValue(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    @patch(MODELS_PATH + ".moons.MOONMINING_VOLUME_PER_MONTH", 1000000)
    @patch(MODELS_PATH + ".extractions.MOONMINING_REPROCESSING_YIELD", 0.7)
    def test_should_calc_correct_value(self):
        # given
        moon = MoonFactory(create_products=False)
        helpers.generate_market_prices(use_process_pricing=False)
        MoonProductFactory(moon=moon, ore_type_id=EveTypeId.CINNABAR, amount=0.19)
        MoonProductFactory(moon=moon, ore_type_id=EveTypeId.CUBIC_BISTOT, amount=0.23)
        MoonProductFactory(
            moon=moon, ore_type_id=EveTypeId.FLAWLESS_ARKONOR, amount=0.25
        )
        MoonProductFactory(
            moon=moon, ore_type_id=EveTypeId.STABLE_VELDSPAR, amount=0.33
        )
        # when
        result = moon.calc_value()
        # then
        self.assertEqual(result, 84622187.5)

    def test_should_return_zero_if_prices_are_missing(self):
        # given
        moon = MoonFactory()
        # when
        result = moon.calc_value()
        # then
        self.assertEqual(result, 0)


class TestMoonCalcRarityClass(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.ore_type_r0 = EveOreType.objects.get(id=EveTypeId.CUBIC_BISTOT)
        cls.ore_type_r4 = EveOreType.objects.get(id=EveTypeId.BITUMENS)
        cls.ore_type_r8 = EveOreType.objects.get(id=EveTypeId.EUXENITE)
        cls.ore_type_r16 = EveOreType.objects.get(id=EveTypeId.CHROMITE)
        cls.ore_type_r32 = EveOreType.objects.get(id=EveTypeId.CINNABAR)
        cls.ore_type_r64 = EveOreType.objects.get(id=EveTypeId.XENOTIME)

    def test_should_return_R4(self):
        # given
        moon = MoonFactory(create_products=False)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r0, amount=0.23)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r4, amount=0.19)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.R4)

    def test_should_return_R8(self):
        # given
        moon = MoonFactory(create_products=False)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r8, amount=0.25)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r0, amount=0.23)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r4, amount=0.19)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.R8)

    def test_should_return_R16(self):
        # given
        moon = MoonFactory(create_products=False)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r4, amount=0.19)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r16, amount=0.23)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r8, amount=0.25)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.R16)

    def test_should_return_R32(self):
        # given
        moon = MoonFactory(create_products=False)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r16, amount=0.23)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r32, amount=0.19)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r8, amount=0.25)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.R32)

    def test_should_return_R64(self):
        # given
        moon = MoonFactory(create_products=False)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r16, amount=0.23)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r32, amount=0.19)
        MoonProductFactory(moon=moon, ore_type=self.ore_type_r64, amount=0.25)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.R64)

    def test_should_handle_moon_without_products(self):
        # given
        moon = MoonFactory(create_products=False)
        # when
        result = moon.calc_rarity_class()
        # then
        self.assertEqual(result, OreRarityClass.NONE)


class TestMoonProductsSorted(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def setUp(self) -> None:
        helpers.generate_market_prices(use_process_pricing=False)

    def test_should_return_moon_products_in_order(self):
        # given
        moon = MoonFactory()
        # when
        result = moon.products_sorted()
        # then
        ore_types = list(result.values_list("ore_type__name", flat=True))
        self.assertListEqual(["Chromite", "Euxenite", "Xenotime"], ore_types)

    def test_should_handle_products_without_price(self):
        # given
        moon = MoonFactory()
        moon_product = moon.products.first()
        EveMarketPrice.objects.filter(
            eve_type_id=moon_product.ore_type_id
        ).average_price = None
        # when
        result = moon.products_sorted()
        # then
        ore_types = list(result.values_list("ore_type__name", flat=True))
        self.assertListEqual(["Chromite", "Euxenite", "Xenotime"], ore_types)

    def test_should_handle_products_without_amount(self):
        # given
        moon = MoonFactory()
        moon_product = moon.products.first()
        moon_product.amount = 0
        moon_product.save()
        # when
        result = moon.products_sorted()
        # then
        ore_types = list(result.values_list("ore_type__name", flat=True))
        self.assertListEqual(["Chromite", "Euxenite", "Xenotime"], ore_types)

    def test_should_handle_products_without_volume(self):
        # given
        moon = MoonFactory()
        moon_product = moon.products.first()
        volume_backup = moon_product.ore_type.volume
        moon_product.ore_type.volume = None
        moon_product.ore_type.save()
        # when
        result = moon.products_sorted()
        # then
        moon_product.ore_type.volume = volume_backup
        moon_product.ore_type.save()
        ore_types = list(result.values_list("ore_type__name", flat=True))
        self.assertListEqual(["Chromite", "Euxenite", "Xenotime"], ore_types)


class TestMoonOverwriteProducts(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        helpers.generate_market_prices()
        load_allianceauth()

    def test_should_overwrite_existing_estimates(self):
        # given
        moon = MoonFactory()
        extraction = CalculatedExtractionFactory()
        ores = {"45506": 7_683_200, "46676": 9_604_000}
        extraction.products = CalculatedExtractionProduct.create_list_from_dict(ores)
        # when
        result = moon.update_products_from_calculated_extraction(extraction)
        # then
        self.assertTrue(result)
        self.assertAlmostEqual(
            moon.products.get(ore_type_id=45506).amount, 0.4, places=2
        )
        self.assertAlmostEqual(
            moon.products.get(ore_type_id=46676).amount, 0.5, places=2
        )
        self.assertIsNone(moon.products_updated_by)
        self.assertIsNotNone(moon.products_updated_at)
        self.assertAlmostEqual(
            moon.products_updated_at, now(), delta=dt.timedelta(minutes=1)
        )

    def test_should_not_overwrite_existing_survey(self):
        # given
        moon = MoonFactory(products_updated_by=UserFactory())
        extraction = CalculatedExtractionFactory()
        ores = {"45506": 7_683_200, "46676": 9_604_000}
        extraction.products = CalculatedExtractionProduct.create_list_from_dict(ores)
        # when
        result = moon.update_products_from_calculated_extraction(extraction)
        # then
        self.assertFalse(result)

    def test_should_overwrite_existing_survey_when_requested(self):
        # given
        moon = MoonFactory(products_updated_by=UserFactory())
        extraction = CalculatedExtractionFactory()
        ores = {"45506": 7_683_200, "46676": 9_604_000}
        extraction.products = CalculatedExtractionProduct.create_list_from_dict(ores)
        # when
        result = moon.update_products_from_calculated_extraction(
            extraction, overwrite_survey=True
        )
        # then
        self.assertTrue(result)

    def test_should_not_overwrite_from_calculated_extraction_without_products(self):
        # given
        moon = MoonFactory()
        extraction = CalculatedExtractionFactory(products=[])
        # when
        result = moon.update_products_from_calculated_extraction(extraction)
        # then
        self.assertFalse(result)
        self.assertTrue(moon.products.exists())

    def test_should_overwrite_products_from_latest_extraction(self):
        # given
        moon = MoonFactory()
        refinery = RefineryFactory(moon=moon)
        ExtractionFactory(refinery=refinery)
        moon.products.all().delete()
        # when
        moon.update_products_from_latest_extraction()
        # then
        self.assertGreater(moon.products.count(), 0)



"""Report views."""

import datetime as dt

from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.models import User
from django.db.models import Count, ExpressionWrapper, F, FloatField, Min, Q, Sum
from django.db.models.functions import Coalesce
from django.http import JsonResponse
from django.shortcuts import render
from django.urls import reverse
from django.utils.html import format_html, strip_tags
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from moonmining.app_settings import (
    MOONMINING_REPROCESSING_YIELD,
    MOONMINING_USE_REPROCESS_PRICING,
    MOONMINING_VOLUME_PER_MONTH,
)
from moonmining.constants import EveGroupId
from moonmining.models import EveOreType, Moon, Refinery


def _previous_month(obj: dt.datetime) -> dt.datetime:
    first = obj.replace(day=1)
    return first - dt.timedelta(days=1)


@login_required()
@permission_required(["moonmining.basic_access", "moonmining.reports_access"])
def reports(request):
    """Render reports view."""
    month_minus_1 = _previous_month(now())
    month_minus_2 = _previous_month(month_minus_1)
    month_minus_3 = _previous_month(month_minus_2)
    month_format = "%b '%y"
    if (
        Refinery.objects.filter(
            owner__is_enabled=True, ledger_last_update_at__isnull=False
        )
        .exclude(ledger_last_update_ok=True)
        .exists()
    ):
        ledger_last_updated = None
    else:
        try:
            ledger_last_updated = Refinery.objects.filter(
                owner__is_enabled=True
            ).aggregate(Min("ledger_last_update_at"))["ledger_last_update_at__min"]
        except KeyError:
            ledger_last_updated = None
    context = {
        "page_title": _("Reports"),
        "use_reprocess_pricing": MOONMINING_USE_REPROCESS_PRICING,
        "reprocessing_yield": MOONMINING_REPROCESSING_YIELD * 100,
        "total_volume_per_month": MOONMINING_VOLUME_PER_MONTH / 1000000,
        "month_minus_3": month_minus_3.strftime(month_format),
        "month_minus_2": month_minus_2.strftime(month_format),
        "month_minus_1": month_minus_1.strftime(month_format),
        "month_current": now().strftime(month_format),
        "ledger_last_updated": ledger_last_updated,
    }
    return render(request, "moonmining/reports.html", context)


def _moon_link_html(moon: Moon) -> str:
    return format_html(
        '<a href="#" data-bs-toggle="modal" '
        'data-bs-target="#modalMoonDetails" '
        'title="{}" '
        "data-ajax_url={}>"
        "{}</a>",
        _("Show details for this moon."),
        reverse("moonmining:moon_details", args=[moon.pk]),
        moon.name,
    )


def _default_if_none(value, default):
    """Return given default if value is None"""
    if value is None:
        return default
    return value


@login_required()
@permission_required(["moonmining.basic_access", "moonmining.reports_access"])
def report_owned_value_data(request):
    """Render data view for owner value report."""
    moon_query = Moon.objects.select_related(
        "eve_moon",
        "eve_moon__eve_planet__eve_solar_system",
        "eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
        "refinery",
        "refinery__owner",
        "refinery__owner__corporation",
        "refinery__owner__corporation__alliance",
    ).filter(refinery__isnull=False)
    corporation_moons = {}
    for moon in moon_query.order_by("eve_moon__name"):
        corporation_name = moon.refinery.owner.name
        if corporation_name not in corporation_moons:
            corporation_moons[corporation_name] = {"moons": [], "total": 0}
        corporation_moons[corporation_name]["moons"].append(moon)
        corporation_moons[corporation_name]["total"] += _default_if_none(moon.value, 0)

    moon_ranks = {
        moon_pk: rank
        for rank, moon_pk in enumerate(
            moon_query.filter(value__isnull=False)
            .order_by("-value")
            .values_list("pk", flat=True)
        )
    }
    grand_total = sum(
        corporation["total"] for corporation in corporation_moons.values()
    )
    data = []
    for corporation_name, details in corporation_moons.items():
        corporation = f"{corporation_name} ({len(details['moons'])})"
        counter = 0
        for moon in details["moons"]:
            grand_total_percent = (
                _default_if_none(moon.value, 0) / grand_total * 100
                if grand_total > 0
                else None
            )
            rank = moon_ranks[moon.pk] + 1 if moon.pk in moon_ranks else None
            data.append(
                {
                    "corporation": corporation,
                    "moon": {"display": _moon_link_html(moon), "sort": counter},
                    "region": moon.region().name,
                    "rarity_class": moon.rarity_tag_html,
                    "value": moon.value,
                    "rank": rank,
                    "total": None,
                    "is_total": False,
                    "grand_total_percent": grand_total_percent,
                }
            )
            counter += 1
        data.append(
            {
                "corporation": corporation,
                "moon": {"display": _("Total"), "sort": counter},
                "region": None,
                "rarity_class": None,
                "value": None,
                "rank": None,
                "total": details["total"],
                "is_total": True,
                "grand_total_percent": None,
            }
        )
    return JsonResponse(data, safe=False)


def _default_if_false(value, default):
    """Return given default if value is False"""
    if not value:
        return default
    return value


@login_required()
@permission_required(["moonmining.basic_access", "moonmining.reports_access"])
def report_user_mining_data(request):
    """Render data view for user mining report."""
    sum_volume = ExpressionWrapper(
        F("mining_ledger__quantity") * F("mining_ledger__ore_type__volume"),
        output_field=FloatField(),
    )
    sum_price = ExpressionWrapper(
        F("mining_ledger__quantity")
        * Coalesce(F("mining_ledger__ore_type__extras__current_price"), 0),
        output_field=FloatField(),
    )
    today = now()
    months_1 = today.replace(day=1) - dt.timedelta(days=1)
    months_2 = months_1.replace(day=1) - dt.timedelta(days=1)
    months_3 = months_2.replace(day=1) - dt.timedelta(days=1)
    users_mining_totals = (
        User.objects.filter(profile__main_character__isnull=False)
        .select_related("profile__main_character", "profile__state")
        .annotate(
            volume_month_0=Sum(
                sum_volume,
                filter=Q(
                    mining_ledger__day__month=today.month,
                    mining_ledger__day__year=today.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            volume_month_1=Sum(
                sum_volume,
                filter=Q(
                    mining_ledger__day__month=months_1.month,
                    mining_ledger__day__year=months_1.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            volume_month_2=Sum(
                sum_volume,
                filter=Q(
                    mining_ledger__day__month=months_2.month,
                    mining_ledger__day__year=months_2.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            volume_month_3=Sum(
                sum_volume,
                filter=Q(
                    mining_ledger__day__month=months_3.month,
                    mining_ledger__day__year=months_3.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            price_month_0=Sum(
                sum_price,
                filter=Q(
                    mining_ledger__day__month=today.month,
                    mining_ledger__day__year=today.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            price_month_1=Sum(
                sum_price,
                filter=Q(
                    mining_ledger__day__month=months_1.month,
                    mining_ledger__day__year=months_1.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            price_month_2=Sum(
                sum_price,
                filter=Q(
                    mining_ledger__day__month=months_2.month,
                    mining_ledger__day__year=months_2.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            price_month_3=Sum(
                sum_price,
                filter=Q(
                    mining_ledger__day__month=months_3.month,
                    mining_ledger__day__year=months_3.year,
                ),
                distinct=True,
            )
        )
    )
    data = []
    for user in users_mining_totals:
        corporation_name = user.profile.main_character.corporation_name
        if user.profile.main_character.alliance_ticker:
            corporation_name += f" [{user.profile.main_character.alliance_ticker}]"
        if any(
            [
                user.volume_month_0,
                user.volume_month_1,
                user.volume_month_2,
                user.volume_month_3,
            ]
        ):
            data.append(
                {
                    "id": user.id,
                    "name": str(user.profile.main_character),
                    "corporation": corporation_name,
                    "state": str(user.profile.state),
                    "volume_month_0": _default_if_false(user.volume_month_0, 0),
                    "volume_month_1": _default_if_false(user.volume_month_1, 0),
                    "volume_month_2": _default_if_false(user.volume_month_2, 0),
                    "volume_month_3": _default_if_false(user.volume_month_3, 0),
                    "price_month_0": _default_if_false(user.price_month_0, 0),
                    "price_month_1": _default_if_false(user.price_month_1, 0),
                    "price_month_2": _default_if_false(user.price_month_2, 0),
                    "price_month_3": _default_if_false(user.price_month_3, 0),
                }
            )
    return JsonResponse(data, safe=False)


@login_required()
@permission_required(["moonmining.basic_access", "moonmining.reports_access"])
def report_user_uploaded_data(request) -> JsonResponse:
    """Render data view for user upload report."""
    data = list(
        Moon.objects.values(
            name=F("products_updated_by__profile__main_character__character_name"),
            corporation=F(
                "products_updated_by__profile__main_character__corporation_name"
            ),
            state=F("products_updated_by__profile__state__name"),
        ).annotate(num_moons=Count("eve_moon_id"))
    )
    for row in data:
        if row["name"] is None:
            row["name"] = "?"
        if row["corporation"] is None:
            row["corporation"] = "?"
        if row["state"] is None:
            row["state"] = "?"
    return JsonResponse(data, safe=False)


@login_required()
@permission_required(["moonmining.basic_access", "moonmining.reports_access"])
def report_ore_prices_data(request) -> JsonResponse:
    """Render data view for ore prices report."""
    moon_ore_group_ids = [
        EveGroupId.UNCOMMON_MOON_ASTEROIDS,
        EveGroupId.UBIQUITOUS_MOON_ASTEROIDS,
        EveGroupId.EXCEPTIONAL_MOON_ASTEROIDS,
        EveGroupId.COMMON_MOON_ASTEROIDS,
        EveGroupId.RARE_MOON_ASTEROIDS,
    ]
    qs = (
        EveOreType.objects.filter(
            eve_group_id__in=moon_ore_group_ids,
            published=True,
            extras__isnull=False,
            extras__current_price__isnull=False,
        )
        .exclude(name__icontains=" ")
        .select_related("eve_group", "extras")
    )
    data = [
        {
            "id": obj.id,
            "name": obj.name,
            "description": strip_tags(obj.description),
            "price": obj.extras.current_price,
            "group": obj.eve_group.name,
            "rarity_html": {
                "display": obj.rarity_class.bootstrap_tag_html,
                "sort": obj.rarity_class.label,
            },
            "rarity_str": obj.rarity_class.label,
        }
        for obj in qs
    ]
    return JsonResponse(data, safe=False)



from app_utils.testing import NoSocketsTestCase

from moonmining.core import CalculatedExtraction
from moonmining.models import Extraction, NotificationType
from moonmining.tests.testdata.factories import (
    ExtractionFactory,
    NotificationFactory,
    RefineryFactory,
)
from moonmining.tests.testdata.load_allianceauth import load_allianceauth
from moonmining.tests.testdata.load_eveuniverse import load_eveuniverse


class TestNotification(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()

    def test_should_convert_to_calculated_extraction(self):
        # given
        refinery = RefineryFactory()
        my_map = [
            (Extraction.Status.STARTED, CalculatedExtraction.Status.STARTED),
            (Extraction.Status.CANCELED, CalculatedExtraction.Status.CANCELED),
            (Extraction.Status.READY, CalculatedExtraction.Status.READY),
            (Extraction.Status.COMPLETED, CalculatedExtraction.Status.COMPLETED),
        ]
        for in_status, out_status in my_map:
            with self.subTest(status=in_status):
                extraction = ExtractionFactory(status=in_status, refinery=refinery)
                notification = NotificationFactory(extraction=extraction)
                # when
                obj = notification.to_calculated_extraction()
                # then
                self.assertEqual(obj.status, out_status)


class TestNotificationType(NoSocketsTestCase):
    def test_str(self):
        # given
        obj = NotificationType.MOONMINING_EXTRACTION_CANCELLED
        # when/then
        self.assertIsInstance(str(obj), str)



"""Helpers."""

import datetime as dt
from collections import defaultdict
from typing import List

from django.utils.html import format_html
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import User


class EnumToDict:
    """Adds ability to an Enum class to be converted to a ordinary dict.

    This e.g. allows using Enums in Django templates.
    """

    @classmethod
    def to_dict(cls) -> dict:
        """Convert this enum to dict."""
        return {k: elem.value for k, elem in cls.__members__.items()}


# pylint: disable = redefined-builtin
def eve_entity_get_or_create_esi_safe(id):
    """Get or Create EveEntity with given ID safely and return it. Else return None."""
    if id:
        try:
            entity, _ = EveEntity.objects.get_or_create_esi(id=id)
            return entity
        except OSError:
            pass
    return None


def round_seconds(obj: dt.datetime) -> dt.datetime:
    """Return copy rounded to full seconds."""
    if obj.microsecond >= 500_000:
        obj += dt.timedelta(seconds=1)
    return obj.replace(microsecond=0)


def user_perms_lookup(user: User, selected_permissions: List[str]) -> dict:
    """Create a lookup for user permissions.

    Allows to create a perms object in Javascript
    that looks like it's namesake in templates.
    """
    all_permissions = user.get_all_permissions()
    user_perms = defaultdict(dict)
    for permission in selected_permissions:
        app_name, perm_name = permission.split(".")
        user_perms[app_name][perm_name] = permission in all_permissions
    return user_perms


def bootstrap5_label_html(text: str, label: str = "default") -> str:
    """Return HTML for a Bootstrap 5 label."""
    return format_html('<span class="badge text-bg-{}">{}</span>', label, text)



"""General models."""

from django.db import models


class General(models.Model):
    """Meta model for global app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access the moonmining app"),
            ("extractions_access", "Can access extractions and view owned moons"),
            ("reports_access", "Can access reports"),
            ("view_all_moons", "Can view all known moons"),
            ("upload_moon_scan", "Can upload moon scans"),
            ("add_refinery_owner", "Can add refinery owner"),
            ("view_moon_ledgers", "Can view moon ledgers"),
        )



"""Extraction models."""

import datetime as dt
from typing import Iterable, Optional

from django.core.exceptions import ObjectDoesNotExist
from django.core.validators import MinValueValidator
from django.db import models
from django.db.models import F, Sum
from django.db.models.functions import Coalesce
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveEntity, EveType

from app_utils.views import BootstrapStyleBS5

from moonmining.app_settings import MOONMINING_REPROCESSING_YIELD
from moonmining.constants import EveDogmaAttributeId, EveGroupId
from moonmining.core import CalculatedExtraction, CalculatedExtractionProduct
from moonmining.helpers import bootstrap5_label_html
from moonmining.managers import EveOreTypeManger, ExtractionManager
from moonmining.models.notifications import NotificationType


class OreQualityClass(models.TextChoices):
    """Quality class of an ore"""

    UNDEFINED = "UN", _("undefined")
    REGULAR = "RE", _("regular")
    IMPROVED = "IM", _("improved")
    EXCELLENT = "EX", _("excellent")

    @property
    def bootstrap_tag_html(self) -> str:
        """Return bootstrap tag."""
        map_quality_to_label_def = {
            self.IMPROVED: {"text": "+15%", "label": BootstrapStyleBS5.SUCCESS},
            self.EXCELLENT: {"text": "+100%", "label": BootstrapStyleBS5.WARNING},
        }
        try:
            label_def = map_quality_to_label_def[self]
            return bootstrap5_label_html(label_def["text"], label=label_def["label"])
        except KeyError:
            return ""

    @classmethod
    def from_eve_type(cls, eve_type: EveType) -> "OreQualityClass":
        """Create object from given eve type."""
        map_value_2_quality_class = {
            1: cls.REGULAR,
            3: cls.IMPROVED,
            5: cls.EXCELLENT,
        }
        try:
            dogma_attribute = eve_type.dogma_attributes.get(
                eve_dogma_attribute_id=EveDogmaAttributeId.ORE_QUALITY
            )
        except ObjectDoesNotExist:
            return cls.UNDEFINED
        try:
            return map_value_2_quality_class[int(dogma_attribute.value)]
        except KeyError:
            return cls.UNDEFINED


class Extraction(models.Model):
    """A mining extraction."""

    class Status(models.TextChoices):
        """An extraction status."""

        STARTED = "ST", _("started")  # has been started
        CANCELED = "CN", _("canceled")  # has been canceled
        READY = "RD", _("ready")  # has finished extraction and is ready to be fractured
        COMPLETED = "CP", _("completed")  # has been fractured
        UNDEFINED = "UN", _("undefined")  # unclear status

        @property
        def bootstrap_tag_html(self) -> str:
            """Return HTML to render as bootstrap tag."""
            map_to_type = {
                self.STARTED: BootstrapStyleBS5.SUCCESS,
                self.CANCELED: BootstrapStyleBS5.DANGER,
                self.READY: BootstrapStyleBS5.WARNING,
                self.COMPLETED: BootstrapStyleBS5.PRIMARY,
                self.UNDEFINED: "",
            }
            try:
                return bootstrap5_label_html(self.label, label=map_to_type[self].value)
            except KeyError:
                return ""

        @property
        def to_notification_type(self) -> NotificationType:
            """Return notification type."""
            map_to_type = {
                self.STARTED: NotificationType.MOONMINING_EXTRACTION_STARTED,
                self.CANCELED: NotificationType.MOONMINING_EXTRACTION_CANCELLED,
                self.READY: NotificationType.MOONMINING_EXTRACTION_FINISHED,
                self.COMPLETED: NotificationType.MOONMINING_LASER_FIRED,
            }
            try:
                return map_to_type[self]
            except KeyError:
                raise ValueError("Invalid status for notification type") from None

        @classmethod
        def considered_active(cls):
            """Return enums considered active."""
            return [cls.STARTED, cls.READY]

        @classmethod
        def considered_inactive(cls):
            """Return enums considered inactive."""
            return [cls.CANCELED, cls.COMPLETED]

        @classmethod
        def from_calculated(cls, calculated: CalculatedExtraction):
            """Create new eum from calculated status."""
            map_from_calculated = {
                CalculatedExtraction.Status.STARTED: cls.STARTED,
                CalculatedExtraction.Status.CANCELED: cls.CANCELED,
                CalculatedExtraction.Status.READY: cls.READY,
                CalculatedExtraction.Status.COMPLETED: cls.COMPLETED,
                CalculatedExtraction.Status.UNDEFINED: cls.UNDEFINED,
            }
            try:
                return map_from_calculated[calculated.status]
            except KeyError:
                return cls.UNDEFINED

    # PK
    refinery = models.ForeignKey(
        "Refinery", on_delete=models.CASCADE, related_name="extractions"
    )
    started_at = models.DateTimeField(
        db_index=True, help_text=_("When this extraction was started")
    )
    # normal properties
    auto_fracture_at = models.DateTimeField(
        help_text=_("When this extraction will be automatically fractured"),
    )
    canceled_at = models.DateTimeField(
        null=True, default=None, help_text=_("When this extraction was canceled")
    )
    canceled_by = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        related_name="+",
        help_text=_("Eve character who canceled this extraction"),
    )
    chunk_arrival_at = models.DateTimeField(
        db_index=True, help_text=_("When this extraction is ready to be fractured")
    )
    fractured_at = models.DateTimeField(
        null=True, default=None, help_text=_("When this extraction was fractured")
    )
    fractured_by = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        related_name="+",
        help_text=_("Eve character who fractured this extraction (if any)"),
    )
    is_jackpot = models.BooleanField(
        default=None,
        null=True,
        help_text=_("Whether this is a jackpot extraction (calculated)"),
    )
    started_by = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        null=True,
        default=None,
        related_name="+",
        help_text=_("Eve character who started this extraction"),
    )
    status = models.CharField(
        max_length=2, choices=Status.choices, default=Status.UNDEFINED, db_index=True
    )
    value = models.FloatField(
        null=True,
        default=None,
        validators=[MinValueValidator(0.0)],
        help_text=_("Estimated value of this extraction (calculated)"),
    )

    objects = ExtractionManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["refinery", "started_at"], name="functional_pk_extraction"
            )
        ]
        verbose_name = _("extraction")
        verbose_name_plural = _("extractions")

    def __str__(self) -> str:
        return f"{self.refinery} - {self.started_at} - {self.status}"

    @property
    def duration(self) -> dt.timedelta:
        """Duration of this extraction."""
        return self.chunk_arrival_at - self.started_at

    @property
    def duration_in_days(self) -> float:
        """Duration of this extraction in days."""
        return self.duration.total_seconds() / (60 * 60 * 24)

    @property
    def status_enum(self) -> "Extraction.Status":
        """Return current status as enum type."""
        return self.Status(self.status)

    def products_sorted(self) -> models.QuerySet:
        """Return current products."""
        try:
            return (
                self.products.select_related(
                    "ore_type", "ore_type__eve_group", "ore_type__extras"
                )
                .annotate(total_price=self._total_price_db_func())
                .order_by("ore_type__name")
            )
        except (ObjectDoesNotExist, AttributeError):
            return type(self).objects.none()

    @cached_property
    def ledger(self) -> models.QuerySet:
        """Return ledger for this extraction."""
        max_day = self.chunk_arrival_at + dt.timedelta(days=6)
        return self.refinery.mining_ledger.filter(
            day__gte=self.chunk_arrival_at,
            day__lte=max_day,
        )

    def calc_value(self) -> Optional[float]:
        """Calculate value estimate."""
        try:
            return self.products.select_related(
                "ore_type", "ore_type__extras"
            ).aggregate(total_price=self._total_price_db_func())["total_price"]
        except (ObjectDoesNotExist, KeyError, AttributeError):
            return None

    @staticmethod
    def _total_price_db_func():
        return Sum(
            Coalesce(F("ore_type__extras__current_price"), 0.0)
            * F("volume")
            / F("ore_type__volume"),
            output_field=models.FloatField(),
        )

    def calc_is_jackpot(self) -> Optional[bool]:
        """Calculate if extraction is jackpot and return result.
        Return None if extraction has no products.
        """
        try:
            products_qualities = [
                product.ore_type.quality_class == OreQualityClass.EXCELLENT
                for product in self.products.select_related("ore_type").all()
            ]
        except (ObjectDoesNotExist, AttributeError):
            return None

        if not products_qualities:
            return None
        return all(products_qualities)

    def update_calculated_properties(self) -> None:
        """Update calculated properties for this extraction."""
        self.value = self.calc_value()
        self.is_jackpot = self.calc_is_jackpot()
        self.save()

    def to_calculated_extraction(self) -> CalculatedExtraction:
        """Generate a calculated extraction from this extraction."""

        def _products_to_calculated_products():
            return [
                CalculatedExtractionProduct(
                    ore_type_id=obj.ore_type_id, volume=obj.volume
                )
                for obj in self.products.all()
            ]

        params = {"refinery_id": self.refinery_id}
        if self.status == self.Status.STARTED:
            params.update(
                {
                    "status": CalculatedExtraction.Status.STARTED,
                    "chunk_arrival_at": self.chunk_arrival_at,
                    "auto_fracture_at": self.auto_fracture_at,
                    "started_at": self.started_at,
                    "started_by": self.started_by,
                    "products": _products_to_calculated_products(),
                }
            )
        elif self.status == self.Status.READY:
            params.update(
                {
                    "status": CalculatedExtraction.Status.READY,
                    "auto_fracture_at": self.auto_fracture_at,
                    "products": _products_to_calculated_products(),
                }
            )
        elif self.status == self.Status.COMPLETED:
            params.update(
                {
                    "fractured_by": self.fractured_by,
                    "fractured_at": self.fractured_at,
                    "status": CalculatedExtraction.Status.COMPLETED,
                    "products": _products_to_calculated_products(),
                }
            )
        elif self.status == self.Status.CANCELED:
            params.update(
                {
                    "status": CalculatedExtraction.Status.CANCELED,
                    "canceled_at": self.canceled_at,
                    "canceled_by": self.canceled_by,
                }
            )
        return CalculatedExtraction(**params)


class OreRarityClass(models.IntegerChoices):
    """Rarity class of an ore"""

    NONE = 0, ""
    R4 = 4, _("R 4")
    R8 = 8, _("R 8")
    R16 = 16, _("R16")
    R32 = 32, _("R32")
    R64 = 64, _("R64")

    @property
    def bootstrap_tag_html(self) -> str:
        """Return as bootstrap tag HTML."""
        map_rarity_to_type = {
            self.R4: BootstrapStyleBS5.PRIMARY,
            self.R8: BootstrapStyleBS5.INFO,
            self.R16: BootstrapStyleBS5.SUCCESS,
            self.R32: BootstrapStyleBS5.WARNING,
            self.R64: BootstrapStyleBS5.DANGER,
        }
        try:
            return bootstrap5_label_html(
                f"R{self.value}", label=map_rarity_to_type[self].value
            )
        except KeyError:
            return ""

    @classmethod
    def from_eve_group_id(cls, eve_group_id: int) -> "OreRarityClass":
        """Create object from eve group ID"""
        map_group_2_rarity = {
            EveGroupId.UBIQUITOUS_MOON_ASTEROIDS.value: cls.R4,
            EveGroupId.COMMON_MOON_ASTEROIDS.value: cls.R8,
            EveGroupId.UNCOMMON_MOON_ASTEROIDS.value: cls.R16,
            EveGroupId.RARE_MOON_ASTEROIDS.value: cls.R32,
            EveGroupId.EXCEPTIONAL_MOON_ASTEROIDS.value: cls.R64,
        }
        try:
            return map_group_2_rarity[eve_group_id]
        except KeyError:
            return cls.NONE

    @classmethod
    def from_eve_type(cls, eve_type: EveType) -> "OreRarityClass":
        """Create object from eve type"""
        return cls.from_eve_group_id(eve_type.eve_group_id)


class EveOreType(EveType):
    """Subset of EveType for all ore types.

    Ensures TYPE_MATERIALS and DOGMAS is always enabled and allows adding methods to types.
    """

    class Meta:
        proxy = True
        verbose_name = _("ore type")
        verbose_name_plural = _("ore types")

    objects = EveOreTypeManger()

    @property
    def icon_url_32(self) -> str:
        """Return icon URL with 32 pixel width."""
        return self.icon_url(32)

    @property
    def rarity_class(self) -> OreRarityClass:
        """Return rarity class."""
        return OreRarityClass.from_eve_type(self)

    @cached_property
    def quality_class(self) -> OreQualityClass:
        """Return quality class."""
        return OreQualityClass.from_eve_type(self)

    @cached_property
    def price(self) -> float:
        """Return calculated price estimate in ISK per unit."""
        result = self.extras.current_price
        return result if result is not None else 0.0

    def price_by_volume(self, volume: int) -> Optional[float]:
        """Return calculated price estimate in ISK for volume in m3."""
        return self.price_by_units(int(volume // self.volume)) if self.volume else None

    def price_by_units(self, units: int) -> float:
        """Return calculated price estimate in ISK for units."""
        return self.price * units

    def calc_refined_value_per_unit(
        self, reprocessing_yield: Optional[float] = None
    ) -> float:
        """Calculate the refined total value per unit and return it."""
        if not reprocessing_yield:
            reprocessing_yield = MOONMINING_REPROCESSING_YIELD
        units = 10000
        r_units = units / 100
        value = 0
        for type_material in self.materials.select_related(
            "material_eve_type__market_price"
        ):
            try:
                price = type_material.material_eve_type.market_price.average_price
            except (ObjectDoesNotExist, AttributeError):
                continue
            if price:
                value += price * type_material.quantity * r_units * reprocessing_yield
        return value / units

    @classmethod
    def _enabled_sections_union(cls, enabled_sections: Iterable[str]) -> set:
        """Return enabled sections with TYPE_MATERIALS and DOGMAS always enabled."""
        enabled_sections = super()._enabled_sections_union(
            enabled_sections=enabled_sections
        )
        enabled_sections.add(cls.Section.TYPE_MATERIALS)
        enabled_sections.add(cls.Section.DOGMAS)
        return enabled_sections


class ExtractionProduct(models.Model):
    """A product within a mining extraction."""

    extraction = models.ForeignKey(
        Extraction, on_delete=models.CASCADE, related_name="products"
    )
    ore_type = models.ForeignKey(EveOreType, on_delete=models.CASCADE, related_name="+")

    volume = models.FloatField(validators=[MinValueValidator(0.0)])

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["extraction", "ore_type"],
                name="functional_pk_extractionproduct",
            )
        ]
        verbose_name = _("extraction product")
        verbose_name_plural = _("extractions products")

    def __str__(self) -> str:
        return f"{self.extraction} - {self.ore_type}"


class EveOreTypeExtras(models.Model):
    """Extra fields for an EveOreType, e.g. for pricing calculations."""

    class PricingMethod(models.TextChoices):
        """A pricing method."""

        UNKNOWN = "UN", _("Undefined")
        EVE_CLIENT = "EC", _("Eve client")
        REPROCESSED_MATERIALS = "RP", _("Reprocessed materials")

    ore_type = models.OneToOneField(
        EveOreType, on_delete=models.CASCADE, related_name="extras"
    )
    current_price = models.FloatField(
        default=None,
        null=True,
        help_text=_("Price used by all price calculations with this type"),
    )
    pricing_method = models.CharField(
        max_length=2, choices=PricingMethod.choices, default=PricingMethod.UNKNOWN
    )

    class Meta:
        verbose_name = _("ore type extra")
        verbose_name_plural = _("ore type extras")

    def __str__(self) -> str:
        return str(self.ore_type)



# Generated by Django 4.0.10 on 2023-04-24 15:18

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("authentication", "0020_userprofile_language_userprofile_night_mode"),
        ("eveuniverse", "0007_evetype_description"),
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("moonmining", "0006_add_labels"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="eveoretype",
            options={"verbose_name": "ore type", "verbose_name_plural": "ore types"},
        ),
        migrations.AlterModelOptions(
            name="eveoretypeextras",
            options={
                "verbose_name": "ore type extra",
                "verbose_name_plural": "ore type extras",
            },
        ),
        migrations.AlterModelOptions(
            name="extraction",
            options={
                "verbose_name": "extraction",
                "verbose_name_plural": "extractions",
            },
        ),
        migrations.AlterModelOptions(
            name="extractionproduct",
            options={
                "verbose_name": "extraction product",
                "verbose_name_plural": "extractions products",
            },
        ),
        migrations.AlterModelOptions(
            name="label",
            options={"verbose_name": "label", "verbose_name_plural": "labels"},
        ),
        migrations.AlterModelOptions(
            name="miningledgerrecord",
            options={
                "verbose_name": "ledger record",
                "verbose_name_plural": "ledger records",
            },
        ),
        migrations.AlterModelOptions(
            name="moon",
            options={"verbose_name": "moon", "verbose_name_plural": "moons"},
        ),
        migrations.AlterModelOptions(
            name="moonproduct",
            options={
                "verbose_name": "moon product",
                "verbose_name_plural": "moons products",
            },
        ),
        migrations.AlterModelOptions(
            name="notification",
            options={
                "verbose_name": "notification",
                "verbose_name_plural": "notifications",
            },
        ),
        migrations.AlterModelOptions(
            name="owner",
            options={"verbose_name": "owner", "verbose_name_plural": "owners"},
        ),
        migrations.AlterModelOptions(
            name="refinery",
            options={"verbose_name": "refinery", "verbose_name_plural": "refineries"},
        ),
        migrations.AlterField(
            model_name="eveoretypeextras",
            name="current_price",
            field=models.FloatField(
                default=None,
                help_text="Price used by all price calculations with this type",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="eveoretypeextras",
            name="pricing_method",
            field=models.CharField(
                choices=[
                    ("UN", "Undefined"),
                    ("EC", "Eve client"),
                    ("RP", "Reprocessed materials"),
                ],
                default="UN",
                max_length=2,
            ),
        ),
        migrations.AlterField(
            model_name="extraction",
            name="auto_fracture_at",
            field=models.DateTimeField(
                help_text="When this extraction will be automatically fractured"
            ),
        ),
        migrations.AlterField(
            model_name="extraction",
            name="canceled_at",
            field=models.DateTimeField(
                default=None, help_text="When this extraction was canceled", null=True
            ),
        ),
        migrations.AlterField(
            model_name="extraction",
            name="chunk_arrival_at",
            field=models.DateTimeField(
                db_index=True, help_text="When this extraction is ready to be fractured"
            ),
        ),
        migrations.AlterField(
            model_name="extraction",
            name="fractured_at",
            field=models.DateTimeField(
                default=None, help_text="When this extraction was fractured", null=True
            ),
        ),
        migrations.AlterField(
            model_name="extraction",
            name="started_at",
            field=models.DateTimeField(
                db_index=True, help_text="When this extraction was started"
            ),
        ),
        migrations.AlterField(
            model_name="moon",
            name="eve_moon",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                primary_key=True,
                related_name="+",
                serialize=False,
                to="eveuniverse.evemoon",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="character_ownership",
            field=models.ForeignKey(
                default=None,
                help_text="Character used to sync this corporation from ESI",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="authentication.characterownership",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="corporation",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                primary_key=True,
                related_name="+",
                serialize=False,
                to="eveonline.evecorporationinfo",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="is_enabled",
            field=models.BooleanField(
                db_index=True,
                default=True,
                help_text="Disabled corporations are excluded from the update process",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="last_update_at",
            field=models.DateTimeField(
                default=None, help_text="Time of last successful update", null=True
            ),
        ),
    ]
